from pm4py import PetriNet

from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import POWL, StrictPartialOrder, DecisionGraph, OperatorPOWL, Operator, SilentTransition

from powl.conversion.to_powl.from_pn.utils.cut_detection import (
    mine_base_case,
    mine_choice_graph,
    mine_partial_order,
    mine_self_loop, mine_skip,
)

from powl.conversion.to_powl.from_pn.utils.preprocessing import (
    preprocess,
    validate_workflow_net, make_self_loop_explicit,
)
from powl.conversion.to_powl.from_pn.utils.subnet_creation import (
    apply_projection, clone_subnet,
)
from powl.conversion.to_powl.from_pn.utils.weak_reachability import (
    get_simplified_reachability_graph,
)


def convert_workflow_net_to_powl(net: PetriNet) -> POWL:
    """
    Convert a Petri net to a POWL model.

    Parameters:
    - net: PetriNet

    Returns:
    - POWL model
    """
    start_place, end_place = validate_workflow_net(net)
    net = preprocess(net)
    res = __translate_petri_to_powl(net, start_place, end_place)
    res = res.reduce_silent_transitions(add_empty_paths=True)
    return res.simplify()


def __translate_petri_to_powl(
    net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place
) -> POWL:

    net, start_place, end_place = make_self_loop_explicit(net, start_place, end_place)

    base_case = mine_base_case(net)
    if base_case:
        return base_case

    # self_loop = mine_self_loop(net, start_place, end_place)
    # if self_loop:
    #     return __translate_self_loop(
    #         net, self_loop[0], self_loop[1], self_loop[2], self_loop[3]
    #     )
    #
    # skip = mine_skip(net, start_place, end_place)
    # if skip:
    #     return __translate_xor(
    #         net, skip[0], skip[1], skip[2]
    #     )

    reachability_map = get_simplified_reachability_graph(net)

    partition = mine_partial_order(net, end_place, reachability_map)
    if len(partition) > 1:
        return __translate_partial_order(net, partition, start_place, end_place)

    partition = mine_choice_graph(net)
    if len(partition) > 1:
        return __translate_choice_graph(net, partition, start_place, end_place)


    raise Exception(
        f"Failed to detected a POWL structure over the following transitions: {net.transitions}"
    )


def __validate_partial_order(po: StrictPartialOrder):
    po.order.add_transitive_edges()
    if po.order.is_irreflexive():
        return po
    else:
        raise Exception("Conversion failed!")


def __translate_to_binary_relation(
    net, transition_groups, i_place: PetriNet.Place, f_place: PetriNet.Place, enforce_unique_connection_points
):

    groups = [tuple(g) for g in transition_groups]
    transition_to_group_map = {transition: g for g in groups for transition in g}

    group_start_places = {g: set() for g in groups}
    group_end_places = {g: set() for g in groups}
    temp_po = BinaryRelation(groups)
    start_groups = set()
    end_groups = set()

    for p in net.places:
        sources = {arc.source for arc in p.in_arcs}
        targets = {arc.target for arc in p.out_arcs}

        # if p is start place and (p -> t), then p should be a start place in the subnet that contains t
        if p == i_place:
            for t in targets:
                group = transition_to_group_map[t]
                group_start_places[group].add(p)
                start_groups.add(group)
        # if p is end place and (t -> p), then p should be end place in the subnet that contains t
        if p == f_place:
            for t in sources:
                group = transition_to_group_map[t]
                group_end_places[group].add(p)
                end_groups.add(group)

        # if (t1 -> p -> t2) and t1 and t2 are in different subsets, then add an edge in the partial order
        # and set p as end place in g1 and as start place in g2
        for t1 in sources:
            group_1 = transition_to_group_map[t1]
            for t2 in targets:
                group_2 = transition_to_group_map[t2]
                if group_1 != group_2:
                    temp_po.add_edge(group_1, group_2)
                    group_end_places[group_1].add(p)
                    group_start_places[group_2].add(p)

    group_to_powl_map = {}
    children = []
    for group in groups:

        subnet, subnet_start_place, subnet_end_place = apply_projection(
            net, set(group), group_start_places[group], group_end_places[group], enforce_unique_connection_points
        )
        child = __translate_petri_to_powl(subnet, subnet_start_place, subnet_end_place)

        group_to_powl_map[group] = child
        children.append(child)

    rel = BinaryRelation(children)
    for source in temp_po.nodes:
        new_source = group_to_powl_map[source]
        for target in temp_po.nodes:
            if temp_po.is_edge(source, target):
                new_target = group_to_powl_map[target]
                rel.add_edge(new_source, new_target)
    return rel, [group_to_powl_map[g] for g in start_groups], [group_to_powl_map[g] for g in end_groups]


def __translate_partial_order(
    net, transition_groups, i_place: PetriNet.Place, f_place: PetriNet.Place
):

    rel, _, _ = __translate_to_binary_relation(net,
                                               transition_groups,
                                               i_place,
                                               f_place,
                                               enforce_unique_connection_points=False)
    po = StrictPartialOrder(rel.nodes)
    po.order = rel
    po = __validate_partial_order(po)
    return po


def __translate_choice_graph(
    net, transition_groups, i_place: PetriNet.Place, f_place: PetriNet.Place
):

    rel, start_nodes, end_nodes = __translate_to_binary_relation(net,
                                                                 transition_groups,
                                                                 i_place,
                                                                 f_place,
                                                                 enforce_unique_connection_points=True)
    cg = DecisionGraph(rel, start_nodes=start_nodes, end_nodes=end_nodes, empty_path=False)
    cg.validate_connectivity()
    return cg


def __create_sub_powl_model(
    net,
    branch: set[PetriNet.Transition],
    start_place: PetriNet.Place,
    end_place: PetriNet.Place,
):
    subnet, subnet_start_place, subnet_end_place = clone_subnet(
        net, branch, start_place, end_place
    )
    powl = __translate_petri_to_powl(subnet, subnet_start_place, subnet_end_place)
    return powl


def __translate_self_loop(
    net: PetriNet,
    do_nodes,
    redo_nodes,
    start_place: PetriNet.Place,
    end_place: PetriNet.Place,
) -> OperatorPOWL:
    do_powl = __create_sub_powl_model(net, do_nodes, start_place, end_place)
    redo_powl = __create_sub_powl_model(net, redo_nodes, end_place, start_place)
    loop_operator = OperatorPOWL(operator=Operator.LOOP, children=[do_powl, redo_powl])
    return loop_operator


def __translate_xor(
    net: PetriNet,
    children,
    start_place: PetriNet.Place,
    end_place: PetriNet.Place,
) -> OperatorPOWL:
    submodel = __create_sub_powl_model(net, children, start_place, end_place)
    xor_operator = OperatorPOWL(operator=Operator.XOR, children=[submodel, SilentTransition()])
    return xor_operator


import pm4py
from pm4py.objects.bpmn.exporter.variants.etree import get_xml_string
from pm4py.objects.bpmn.layout import layouter as bpmn_layouter

from powl.visualization.bpmn.resource_utils.layouter import (
    apply_layouting as bpmn_io_layout,
)


def layout_bpmn(bpmn_model: pm4py.objects.bpmn.obj.BPMN) -> str:
    """
    Apply layout to a BPMN model.

    :param bpmn_model: BPMN model as a string or PM4Py BPMN object
    :return: String representation of the layouted BPMN model
    """
    if isinstance(bpmn_model, str):
        raise ValueError("Input must be a PM4Py BPMN object, not a string.")
    layouted_bpmn = bpmn_layouter.apply(bpmn_model)
    bpmn = get_xml_string(layouted_bpmn).decode("utf-8")
    bpmn = bpmn_io_layout(bpmn)
    return bpmn



from collections import defaultdict, deque
from copy import copy
from typing import TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.objects.dfg.obj import DFG

T = TypeVar("T", bound=IMDataStructureUVCL)


# Hidden artificial nodes (not stored in DFG.graph; only used for reasoning)
_START = "__START__"
_END = "__END__"


def _build_augmented_adjacency(graph, start_activities, end_activities):
    """
    Builds adjacency including hidden START/END:
      START -> a  for a in start_activities
      a -> END    for a in end_activities
      a -> b      for (a,b) in graph
    Returns (out_adj, in_adj, edge_weight_fn)
    """
    out_adj = defaultdict(list)
    in_adj = defaultdict(list)

    # DFG internal edges
    for (a, b), f in graph.items():
        out_adj[a].append(b)
        in_adj[b].append(a)

    # START arcs
    for a in start_activities.keys():
        out_adj[_START].append(a)
        in_adj[a].append(_START)

    # END arcs
    for a in end_activities.keys():
        out_adj[a].append(_END)
        in_adj[_END].append(a)

    def edge_weight(u, v):
        if u == _START:
            return start_activities[v]
        if v == _END:
            return end_activities[u]
        return graph[(u, v)]

    return out_adj, in_adj, edge_weight


def _bfs_reachable(starts, out_adj):
    q = deque(starts)
    seen = set(starts)
    while q:
        x = q.popleft()
        for y in out_adj.get(x, ()):
            if y not in seen:
                seen.add(y)
                q.append(y)
    return seen


def _bfs_parents_multi_source(sources, out_adj):
    """
    Multi-source BFS that records one parent for each visited node.
    parent[src] is None for all sources.
    """
    q = deque(sources)
    parent = {s: None for s in sources}
    while q:
        u = q.popleft()
        for v in out_adj.get(u, ()):
            if v not in parent:
                parent[v] = u
                q.append(v)
    return parent


def _reverse_bfs_next_multi_sink(sinks, in_adj):
    """
    Multi-sink reverse BFS that records one 'next hop' towards some sink.
    next_hop[x] = y means x -> y moves you closer to a sink.
    For sinks themselves, next_hop[sink] = None.
    """
    q = deque(sinks)
    next_hop = {t: None for t in sinks}
    while q:
        v = q.popleft()
        for u in in_adj.get(v, ()):
            if u not in next_hop:
                next_hop[u] = v
                q.append(u)
    return next_hop


def _add_augmented_edge_to_dfg(dfg, u, v, w):
    """
    Writes an augmented edge back into the DFG object:
      START->a => dfg.start_activities[a] = w
      a->END   => dfg.end_activities[a] = w
      a->b     => dfg.graph[(a,b)] = w
    """
    if u == _START:
        # START -> v
        dfg.start_activities[v] = max(dfg.start_activities.get(v, 0), w)
    elif v == _END:
        # u -> END
        dfg.end_activities[u] = max(dfg.end_activities.get(u, 0), w)
    else:
        dfg.graph[(u, v)] = max(dfg.graph.get((u, v), 0), w)


import math
import heapq

def _widest_parents_multi_source(sources, out_adj, edge_weight):
    """
    Multi-source widest-path (maximum bottleneck) that records one parent.
    width[x] = best bottleneck capacity from any source to x.
    parent[src] = None for all sources.
    """
    width = {s: math.inf for s in sources}
    parent = {s: None for s in sources}

    # max-heap via negative keys
    heap = [(-width[s], s) for s in sources]
    heapq.heapify(heap)

    while heap:
        neg_w, u = heapq.heappop(heap)
        w_u = -neg_w
        if w_u != width.get(u, None):
            continue  # stale heap entry

        for v in out_adj.get(u, ()):
            cap = min(w_u, edge_weight(u, v))
            if cap > width.get(v, -1):
                width[v] = cap
                parent[v] = u
                heapq.heappush(heap, (-cap, v))

    return parent


def _reverse_widest_next_multi_sink(sinks, in_adj, edge_weight):
    """
    Multi-sink reverse widest-path that records one 'next hop' towards some sink.
    width[x] = best bottleneck capacity from x to any sink (in forward direction).
    next_hop[x] = y means x -> y is the chosen next step.

    For sinks themselves: next_hop[sink] = None.
    """
    width = {t: math.inf for t in sinks}
    next_hop = {t: None for t in sinks}

    heap = [(-width[t], t) for t in sinks]
    heapq.heapify(heap)

    while heap:
        neg_w, v = heapq.heappop(heap)
        w_v = -neg_w
        if w_v != width.get(v, None):
            continue  # stale

        # go backwards: u -> v (forward), so u is predecessor of v
        for u in in_adj.get(v, ()):
            cap = min(w_v, edge_weight(u, v))
            if cap > width.get(u, -1):
                width[u] = cap
                next_hop[u] = v
                heapq.heappush(heap, (-cap, u))

    return next_hop


def filter_dfg_noise_keep_activities_and_repair(obj, noise_threshold):
    # Original structures
    start0 = copy(obj.dfg.start_activities)
    end0 = copy(obj.dfg.end_activities)
    graph0 = copy(obj.dfg.graph)

    if not end0 or not start0:
        raise ValueError("Assumption violated: no original start or end activities!")

    # Universe of activities from original DFG
    activities = set(start0.keys()) | set(end0.keys())
    for (a, b) in graph0.keys():
        activities.add(a)
        activities.add(b)

    # --- Check the assumption on the ORIGINAL augmented graph:
    out0, in0, w0 = _build_augmented_adjacency(graph0, start0, end0)
    R0 = _bfs_reachable({_START}, out0)
    # Compute nodes that can reach END in original:
    next_to_end0 = _reverse_bfs_next_multi_sink({_END}, in0)
    on_path0 = {a for a in activities if a in R0 and a in next_to_end0}
    if on_path0 != activities:
        missing = activities - on_path0
        raise ValueError(
            f"Assumption violated: {len(missing)} activities are not on any START->END path in the original DFG."
        )

    # --- Noise filter
    outgoing_max_occ = {}
    for (a, b), f in graph0.items():
        outgoing_max_occ[a] = max(outgoing_max_occ.get(a, 0), f)
    for a, f in end0.items():
        outgoing_max_occ[a] = max(outgoing_max_occ.get(a, 0), f)

    # Filter internal edges
    graph1 = {
        (a, b): f
        for (a, b), f in graph0.items()
        if f >= noise_threshold * outgoing_max_occ.get(a, f)
    }

    dfg = DFG()
    dfg.graph.update(graph1)

    # Filter start arcs
    if start0:
        start_max = max(start0.values())
        for a, f in start0.items():
            if f >= noise_threshold * start_max:
                dfg.start_activities[a] = f

    # Filter end arcs
    for a, f in end0.items():
        if f >= noise_threshold * outgoing_max_occ.get(a, f):
            dfg.end_activities[a] = f

    # If we filtered all ends/starts away, that’s OK — repair is allowed to re-add them.
    # But we need at least one end to define END reachability during repair.
    if not dfg.end_activities:
        # pick most frequent original end (deterministic)
        a, f = max(end0.items(), key=lambda kv: kv[1])
        dfg.end_activities[a] = f

    if not dfg.start_activities:
        # pick most frequent original start if any; otherwise repair will re-add via paths anyway
        if start0:
            a, f = max(start0.items(), key=lambda kv: kv[1])
            dfg.start_activities[a] = f
        else:
            # If original start_activities is empty (rare), that contradicts standard DFG semantics
            raise ValueError("Original start_activities is empty; cannot define START.")

    # --- Guaranteed repair using ORIGINAL augmented paths
    while True:
        out_cur, in_cur, w_cur = _build_augmented_adjacency(dfg.graph, dfg.start_activities, dfg.end_activities)

        # current backbone: reachable from START and can reach END
        R = _bfs_reachable({_START}, out_cur)
        next_to_end = _reverse_bfs_next_multi_sink({_END}, in_cur)
        backbone = {a for a in activities if a in R and a in next_to_end}

        broken = activities - backbone
        if not broken:
            break

        # Build “how to get there” from current reachable set, but using ORIGINAL graph (so we can re-add missing arcs)
        # Sources for forward parents: everything currently reachable (including START)
        # Sinks for reverse next: everything currently co-reachable to END (i.e., present in next_to_end)
        # Note: co-reachable set is exactly keys of next_to_end (nodes that can reach END in current)
        co_reachable = set(next_to_end.keys())

        parent_from_R_in_orig = _widest_parents_multi_source(R, out0, w0)
        next_to_C_in_orig = _reverse_widest_next_multi_sink(co_reachable, in0, w0)

        progressed = False

        for a in broken:
            # 1) Add a path from some node in R to a (in original augmented graph)
            if a not in parent_from_R_in_orig:
                # With your assumption + because START is in R, this should not happen
                raise ValueError(f"Repair failed: cannot reach {a} from current reachable set in original graph.")

            # Reconstruct forward path: ... -> a
            path_nodes = []
            x = a
            while x is not None and x not in R:
                path_nodes.append(x)
                x = parent_from_R_in_orig[x]
            # Now x is in R; parent pointers give edges parent->child
            for node in reversed(path_nodes):
                u = parent_from_R_in_orig[node]
                v = node
                if u is None:
                    # node itself is a source; shouldn't occur for nodes not in R
                    continue
                _add_augmented_edge_to_dfg(dfg, u, v, w0(u, v))
                progressed = True

            # 2) Add a path from a to some node that can reach END (co_reachable), in original augmented graph
            if a not in next_to_C_in_orig:
                raise ValueError(f"Repair failed: cannot connect {a} to END from current co-reachable set in original graph.")

            x = a
            while x is not None and x not in co_reachable:
                y = next_to_C_in_orig[x]
                if y is None:
                    break
                _add_augmented_edge_to_dfg(dfg, x, y, w0(x, y))
                progressed = True
                x = y

        if not progressed:
            raise ValueError("Repair made no progress; this should be impossible under the connectivity assumption.")

    return IMDataStructureUVCL(obj.data_structure, dfg)



from abc import ABC
from collections import Counter
from itertools import combinations
from typing import Any, Collection, Dict, Generic, List, Optional, Tuple

from pm4py.algo.discovery.inductive.cuts import utils as cut_util

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG
from pm4py.objects.process_tree.obj import Operator

from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import DecisionGraph, POWL


class MaximalDecisionGraphCut(Cut[T], ABC, Generic[T]):
    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> Operator:
        return None

    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Any]]:

        alphabet = parameters["alphabet"]
        transitive_successors = parameters["transitive_successors"]

        groups = [frozenset([a]) for a in alphabet]

        for a, b in combinations(alphabet, 2):
            if b in transitive_successors[a] and a in transitive_successors[b]:
                groups = cut_util.merge_groups_based_on_activities(a, b, groups)


        if len(groups) < 2:
            return None

        return groups

    @classmethod
    def apply(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[DecisionGraph, List[POWL]]]:

        dfg = obj.dfg

        start_acts = set(obj.dfg.start_activities.keys())
        end_acts = set(obj.dfg.end_activities.keys())

        alphabet = sorted(dfu.get_vertices(dfg), key=lambda g: g.__str__())
        transitive_predecessors, transitive_successors = dfu.get_transitive_relations(
            dfg
        )

        parameters["alphabet"] = alphabet
        parameters["transitive_predecessors"] = transitive_predecessors
        parameters["transitive_successors"] = transitive_successors

        groups = cls.holds(obj, parameters)
        if groups is None:
            return groups

        # cache start/end flags per group
        group_has_start = [any(a in start_acts for a in g) for g in groups]
        group_has_end = [any(a in end_acts for a in g) for g in groups]

        # cache group->group connectivity (directed)
        n = len(groups)
        group_conn = [[False] * n for _ in range(n)]

        # efficient: build act->group index
        act_to_group = {}
        for i, g in enumerate(groups):
            for a in g:
                act_to_group[a] = i

        for (a, b), freq in dfg.graph.items():
            if freq <= 0:
                continue
            ga = act_to_group.get(a)
            gb = act_to_group.get(b)
            if ga != gb:
                group_conn[ga][gb] = True

        parameters["_mdgc_act_to_group"] = act_to_group
        parameters["_mdgc_group_has_start"] = group_has_start
        parameters["_mdgc_group_has_end"] = group_has_end
        parameters["_mdgc_group_conn"] = group_conn

        children = cls.project(obj, groups, parameters)

        order = BinaryRelation(nodes=children)

        for i in range(n):
            for j in range(n):
                if i != j and group_conn[i][j]:
                    order.add_edge(children[i], children[j])

        start_nodes = [children[i] for i in range(n) if group_has_start[i]]
        end_nodes = [children[i] for i in range(n) if group_has_end[i]]

        dg = DecisionGraph(order, start_nodes, end_nodes)
        return dg, dg.children


class MaximalDecisionGraphCutUVCL(MaximalDecisionGraphCut[IMDataStructureUVCL], ABC):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:

        logs = [Counter() for _ in groups]
        for t, freq in obj.data_structure.items():
            for i, group in enumerate(groups):
                seg = []
                for e in t:
                    if e in group:
                        seg.append(e)
                if len(seg) > 0:
                    logs[i][tuple(seg)] += freq

        return [IMDataStructureUVCL(l) for l in logs]


class MaximalDecisionGraphCutDFG(MaximalDecisionGraphCut[IMDataStructureDFG], ABC):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureDFG,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:

        base_dfg = obj.dfg
        dfg_map = {group: DFG() for group in groups}

        activity_to_group_map = {}
        for group in groups:
            for activity in group:
                activity_to_group_map[activity] = group

        for (a, b) in base_dfg.graph:
            group_a = activity_to_group_map[a]
            group_b = activity_to_group_map[b]
            freq = base_dfg.graph[(a, b)]
            if group_a == group_b:
                dfg_map[group_a].graph[(a, b)] = freq
            else:
                dfg_map[group_a].end_activities[a] += freq
                dfg_map[group_b].start_activities[b] += freq
        for a in base_dfg.start_activities:
            group_a = activity_to_group_map[a]
            dfg_map[group_a].start_activities[a] += base_dfg.start_activities[a]
        for a in base_dfg.end_activities:
            group_a = activity_to_group_map[a]
            dfg_map[group_a].end_activities[a] += base_dfg.end_activities[a]

        return list(
            map(
                lambda g: IMDataStructureDFG(InductiveDFG(dfg=dfg_map[g], skip=False)),
                groups,
            )
        )



from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructure,
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg import util as dfu

from powl.discovery.total_order_based.inductive.cuts.concurrency import (
    POWLConcurrencyCutDFG,
    POWLConcurrencyCutUVCL,
)
from powl.discovery.total_order_based.inductive.cuts.factory import CutFactory, T
from powl.discovery.total_order_based.inductive.cuts.loop import (
    POWLLoopCutDFG,
    POWLLoopCutUVCL,
)
from powl.discovery.total_order_based.inductive.variants.decision_graph.max_decision_graph_cut import (
    MaximalDecisionGraphCutDFG,
    MaximalDecisionGraphCutUVCL,
)
from powl.discovery.total_order_based.inductive.variants.maximal.maximal_partial_order_cut import (
    MaximalPartialOrderCutDFG,
    MaximalPartialOrderCutUVCL,
)
from powl.objects.obj import POWL


class CutFactoryPOWLDecisionGraphMaximal(CutFactory):
    @classmethod
    def get_cuts(cls, obj, parameters=None):

        if type(obj) is IMDataStructureUVCL:
            return [
                MaximalDecisionGraphCutUVCL,
                MaximalPartialOrderCutUVCL,
                POWLConcurrencyCutUVCL,
                POWLLoopCutUVCL
            ]
        elif type(obj) is IMDataStructureDFG:
            return [
                MaximalDecisionGraphCutDFG,
                POWLConcurrencyCutDFG,
                POWLLoopCutDFG,
                MaximalPartialOrderCutDFG,
            ]
        else:
            return []

    @classmethod
    def find_cut(
        cls, obj: IMDataStructure, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        alphabet = sorted(dfu.get_vertices(obj.dfg), key=lambda g: g.__str__())
        if len(alphabet) < 2:
            return None
        for c in CutFactoryPOWLDecisionGraphMaximal.get_cuts(obj):
            r = c.apply(obj, parameters)
            if r is not None:
                return r
        return None



from collections import Counter


def get_transitive_closure_from_counter(relation: Counter):
    nodes = set()
    adj = {}

    for (a, b), count in relation.items():
        if count > 0:
            nodes.update([a, b])
            adj.setdefault(a, set()).add(b)

    closure = {u: set(adj.get(u, ())) for u in nodes}
    for k in nodes:
        for i in nodes:
            if k in closure[i]:
                closure[i] |= closure[k]

    return closure



from typing import Any, Dict, Optional

from pm4py.objects.ocel.obj import OCEL

from powl.discovery.dfg_based.algorithm import apply as discover_from_dfg

from powl.discovery.object_centric.variants.oc_powl.utils.divergence_free_graph import (
    get_divergence_free_graph,
)
from powl.discovery.object_centric.variants.oc_powl.utils.filtering import (
    keep_most_frequent_activities,
)
from powl.discovery.object_centric.variants.oc_powl.utils.interaction_properties import (
    get_interaction_patterns,
)
from powl.discovery.object_centric.variants.oc_powl.utils.ocpn_conversion import (
    convert_ocpowl_to_ocpn,
)
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.oc_powl import load_oc_powl


def apply(
    oc_log: OCEL,
    powl_miner_variant=POWLDiscoveryVariant.MAXIMAL,
    activity_coverage_threshold: float = 1.0,
    parameters: Optional[Dict[Any, Any]] = None,
) -> Dict[str, Any]:

    relations = oc_log.relations
    relations = keep_most_frequent_activities(
        relations, coverage=activity_coverage_threshold
    )

    div, con, rel, defi = get_interaction_patterns(relations)
    df2_graph, divergence_matrices = get_divergence_free_graph(relations, div, rel)

    powl_model = discover_from_dfg(df2_graph, variant=powl_miner_variant)
    import powl

    powl.view(powl_model)
    oc_powl = load_oc_powl(powl_model, rel, div, con, defi)
    ocpn = convert_ocpowl_to_ocpn(oc_powl, divergence_matrices)

    return ocpn



from abc import ABC
from typing import Any, Dict, Generic, Optional

from pm4py.algo.discovery.inductive.cuts.xor import (
    ExclusiveChoiceCut,
    ExclusiveChoiceCutDFG,
    ExclusiveChoiceCutUVCL,
    T,
)
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.process_tree.obj import Operator

from powl.objects.obj import OperatorPOWL


class POWLExclusiveChoiceCut(ExclusiveChoiceCut, ABC, Generic[T]):
    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> OperatorPOWL:
        return OperatorPOWL(Operator.XOR, [])


class POWLExclusiveChoiceCutUVCL(
    ExclusiveChoiceCutUVCL, POWLExclusiveChoiceCut[IMDataStructureUVCL], ABC
):
    pass


class POWLExclusiveChoiceCutDFG(
    ExclusiveChoiceCutDFG, POWLExclusiveChoiceCut[IMDataStructureDFG], ABC
):
    pass



from typing import List, Tuple

from powl.visualization.bpmn.resource_utils.lanes import Lane


class Pool:
    def __init__(
        self,
        up_left: Tuple[int, int],
        down_right: Tuple[int, int],
        name: str,
        lanes: List[Lane],
    ):
        self.up_left = up_left
        self.down_right = down_right
        self.name = name
        self.lanes = lanes

    def get_up_left(self) -> Tuple[int, int]:
        return self.up_left

    def set_up_left(self, up_left: Tuple[int, int]) -> None:
        self.up_left = up_left

    def get_down_right(self) -> Tuple[int, int]:
        return self.down_right

    def set_down_right(self, down_right: Tuple[int, int]) -> None:
        self.down_right = down_right

    def get_name(self) -> str:
        return self.name

    def get_lanes(self) -> List[Lane]:
        return self.lanes



from collections import defaultdict

from powl.discovery.partial_order_based.utils.simplified_objects import (
    Graph,
    SelfLoop,
    Skip,
    SkipSelfLoop,
)


def find_self_loops(mapping, new_nodes_counter):
    new_mapping = {}
    reversed_mapping = defaultdict(set)

    for key, value in mapping.items():
        reversed_mapping[value].add(key)

    tagged_node_element_map = defaultdict(list)

    for node in reversed_mapping.keys():
        if (
            isinstance(node, Skip)
            or isinstance(node, SelfLoop)
            or isinstance(node, SkipSelfLoop)
        ):
            pass
        else:
            continue
        tagged_node_element_map[node.element].append(node)

    processed_keys = set()

    for tagged_node_element in tagged_node_element_map.keys():

        element_list = tagged_node_element_map[tagged_node_element]

        if tagged_node_element in reversed_mapping.keys():
            new_node = SelfLoop.create(tagged_node_element)
            values = reversed_mapping[tagged_node_element]
            processed_keys.add(tagged_node_element)
            for tagged_node in element_list:
                values.update(reversed_mapping[tagged_node])
                processed_keys.add(tagged_node)
            for node in values:
                new_mapping[node] = new_node

        elif len(element_list) > 1:
            if any(isinstance(tagged_node, SelfLoop) for tagged_node in element_list):
                new_node = SelfLoop.create(tagged_node_element)
            else:
                new_node = SkipSelfLoop.create(tagged_node_element)

            values = set()
            for tagged_node in element_list:
                values.update(reversed_mapping[tagged_node])
                processed_keys.add(tagged_node)

            for node in values:
                new_mapping[node] = new_node

    for key, value in reversed_mapping.items():
        if key in processed_keys:
            continue
        if new_nodes_counter[key] > 1:
            if isinstance(key, Skip):
                new_node = SkipSelfLoop.create(key.element)
            else:
                new_node = SelfLoop.create(key)
        else:
            new_node = key
        for node in value:
            new_mapping[node] = new_node

    return new_mapping


def apply_node_mapping_on_single_graph(graph: Graph, node_mapping: dict):

    reverse_mapping = defaultdict(set)
    new_nodes = set()
    for n in graph.nodes:
        if n in node_mapping:
            value = node_mapping[n]
            new_nodes.add(value)
            reverse_mapping[value].add(n)
        else:
            raise ValueError

    new_nodes = frozenset(new_nodes)
    new_edges = set()

    for source in new_nodes:
        for target in new_nodes:
            if source != target and all(
                (s, t) in graph.edges
                for s in reverse_mapping[source]
                for t in reverse_mapping[target]
            ):
                new_edges.add((source, target))
    return Graph(
        nodes=new_nodes,
        edges=frozenset(new_edges),
        additional_information=graph.additional_information,
    )



from typing import Any, Dict, List as TList, Optional, Type, TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructure,
    IMDataStructureDFG,
    IMDataStructureUVCL,
)

from powl.discovery.total_order_based.inductive.base_case.abc import BaseCase
from powl.discovery.total_order_based.inductive.base_case.empty_log import (
    EmptyLogBaseCaseDFG,
    EmptyLogBaseCaseUVCL,
)
from powl.discovery.total_order_based.inductive.base_case.single_activity import (
    SingleActivityBaseCaseDFG,
    SingleActivityBaseCaseUVCL,
)

from powl.objects.obj import POWL

T = TypeVar("T", bound=IMDataStructure)
S = TypeVar("S", bound=BaseCase)


class BaseCaseFactory:
    @classmethod
    def get_base_cases(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> TList[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            return [EmptyLogBaseCaseUVCL, SingleActivityBaseCaseUVCL]
        elif type(obj) is IMDataStructureDFG:
            return [EmptyLogBaseCaseDFG, SingleActivityBaseCaseDFG]
        else:
            return []

    @classmethod
    def apply_base_cases(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[POWL]:
        for b in BaseCaseFactory.get_base_cases(obj):
            r = b.apply(obj, parameters)
            if r is not None:
                return r
        return None



from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Optional, TypeVar, Union

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure

from powl.objects.obj import POWL

T = TypeVar("T", bound=Union[IMDataStructure])


class BaseCase(ABC, Generic[T]):
    @classmethod
    def apply(
        cls, obj=T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[POWL]:
        return cls.leaf(obj, parameters) if cls.holds(obj, parameters) else None

    @classmethod
    @abstractmethod
    def holds(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> bool:
        pass

    @classmethod
    @abstractmethod
    def leaf(cls, obj=T, parameters: Optional[Dict[str, Any]] = None) -> POWL:
        pass



from copy import copy

from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils as pn_util

from powl.conversion.to_powl.from_pn.utils.subnet_creation import pn_transition_to_powl, add_arc_from_to, clone_place
from powl.conversion.to_powl.from_pn.utils.weak_reachability import (
    get_reachable_without_looping,
)


def mine_base_case(net: PetriNet):
    if len(net.transitions) == 1 and len(net.places) == 2 == len(net.arcs):
        activity = list(net.transitions)[0]
        powl_transition = pn_transition_to_powl(activity)
        return powl_transition
    return None


def mine_self_loop(
    net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place
):

    if len(start_place.out_arcs) == 1 and len(end_place.in_arcs) == 1:
        t = pn_util.pre_set(end_place)
        if t in pn_util.post_set(start_place) and isinstance(t, PetriNet.Transition) and not t.label:
            return {t}, copy(net.transitions), start_place, end_place

    if len(start_place.in_arcs) == 1 and len(end_place.out_arcs) == 1:
        t = pn_util.pre_set(start_place)
        if t in pn_util.post_set(end_place) and isinstance(t, PetriNet.Transition) and not t.label:
            return copy(net.transitions), {t}, start_place, end_place

    return None


def mine_skip(
    net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place
):

    if len(start_place.in_arcs) == 0 and len(end_place.out_arcs) == 0:
        silent_connectors = [
            t for t in pn_util.pre_set(end_place) if isinstance(t, PetriNet.Transition)
                                                     and not t.label and len(t.out_arcs) == 1
                                                     and len(t.in_arcs) == 1
                                                     and t in pn_util.post_set(start_place)
        ]
        if len(silent_connectors) > 0:
            other_children = {t for t in net.transitions if t not in silent_connectors}
            if len(other_children) > 0:
                return other_children, start_place, end_place

    return None


def mine_partial_order(net, end_place, reachability_map):
    partition = [{t} for t in net.transitions]

    for place in net.places:
        out_size = len(place.out_arcs)
        if out_size > 1 or (place == end_place and out_size > 0):
            xor_branches = []
            for start_transition in pn_util.post_set(place):
                new_branch = {
                    node
                    for node in reachability_map[start_transition]
                    if isinstance(node, PetriNet.Transition)
                }
                xor_branches.append(new_branch)
            union_of_branches = set().union(*xor_branches)
            if place == end_place:
                not_in_every_branch = union_of_branches
            else:
                intersection_of_branches = set.intersection(*xor_branches)
                not_in_every_branch = union_of_branches - intersection_of_branches
            if len(not_in_every_branch) > 1:
                partition = __combine_parts(not_in_every_branch, partition)

    return partition


def mine_choice_graph(net):
    partition = [{t} for t in net.transitions]

    split_transitions = [transition for transition in net.transitions if  len(transition.out_arcs) > 1]
    join_transitions = [transition for transition in net.transitions if len(transition.in_arcs) > 1]

    for split in split_transitions:
        split_branches = []
        for start_place in pn_util.post_set(split):
            new_branch = {
                node
                for node in get_reachable_without_looping(start_place, split)
                if isinstance(node, PetriNet.Transition)
            }
            split_branches.append(new_branch)
        union_of_branches = set().union(*split_branches)
        intersection_of_branches = set.intersection(*split_branches)
        not_in_every_branch = union_of_branches - intersection_of_branches
        if len(not_in_every_branch) > 1:
            not_in_every_branch.add(split)
            partition = __combine_parts(not_in_every_branch, partition)
        else:
            raise Exception("This should not happen!")

    for join in join_transitions:
        pre_transitions = {join}
        for pre_place in pn_util.pre_set(join):
            pre_transitions = pre_transitions | pn_util.pre_set(pre_place)
        partition = __combine_parts(pre_transitions, partition)

    return partition


def __combine_parts(
    transitions_to_group_together: set[PetriNet.Transition],
    partition: list[set[PetriNet.Transition]],
):
    new_partition = []
    new_combined_group = set()

    for part in partition:

        if part & transitions_to_group_together:
            new_combined_group.update(part)
        else:
            new_partition.append(part)

    if new_combined_group:
        new_partition.append(new_combined_group)

    return new_partition


from typing import List

import networkx as nx
import pm4py.objects.bpmn.obj as bpmn_obj
from pm4py.objects.process_tree.obj import Operator

from powl.objects.obj import (
    DecisionGraph,
    EndNode,
    OperatorPOWL,
    SilentTransition,
    StartNode,
    StrictPartialOrder,
    Transition,
)


def __handle_transition(powl_content: Transition) -> nx.DiGraph:
    # Add artificial start and end nodes
    subgraph = nx.DiGraph()
    start_node = f"Start_{id(powl_content)}"
    end_node = f"End_{id(powl_content)}"
    subgraph.add_node(start_node, type="start", visited=True)
    subgraph.add_node(end_node, type="end", visited=True)
    subgraph.add_node(id(powl_content), content=powl_content.label, visited=True)
    # Add the edges start -> transition -> end
    subgraph.add_edge(start_node, id(powl_content))
    subgraph.add_edge(id(powl_content), end_node)
    return subgraph


def __handle_silent_transition(powl_content: SilentTransition) -> nx.DiGraph:
    subgraph = nx.DiGraph()
    start_node = f"Start_{id(powl_content)}"
    end_node = f"End_{id(powl_content)}"
    subgraph.add_node(start_node, type="start", visited=True)
    subgraph.add_node(end_node, type="end", visited=True)
    # Add the edges start -> silent transition -> end
    subgraph.add_node(start_node, visited=True)
    subgraph.add_node(end_node, visited=True)
    subgraph.add_edge(start_node, end_node)
    return subgraph


def __handle_operator_powl(powl_content: OperatorPOWL) -> nx.DiGraph:
    """
    Handle the OperatorPOWL content and return a directed graph.

    Parameters
    ----------
    powl_content : OperatorPOWL
        The OperatorPOWL content to handle.

    Returns
    -------
    G : nx.DiGraph
        The directed graph representing the OperatorPOWL.
    """
    G = nx.DiGraph()
    start_event = f"Start_{id(powl_content)}"
    end_event = f"End_{id(powl_content)}"
    G.add_node(start_event, type="start", visited=True)
    G.add_node(end_event, type="end", visited=True)
    operator = powl_content.operator

    if operator == Operator.LOOP and len(powl_content.children) == 2:
        # Add a diverging parallel gateway and a converging parallel gateway
        diverging_gateway = f"ExclusiveGateway_{id(powl_content)}_diverging"
        converging_gateway = f"ExclusiveGateway_{id(powl_content)}_converging"
        G.add_node(
            diverging_gateway,
            type="diverging",
            paired_with=[converging_gateway],
            visited=True,
        )
        G.add_node(
            converging_gateway,
            type="converging",
            paired_with=[diverging_gateway],
            visited=True,
        )
        # Handle the edges now
        do_part = powl_content.children[0]
        redo_part = powl_content.children[1]
        G.add_node(id(do_part), content=do_part, visited=False)
        G.add_node(id(redo_part), content=redo_part, visited=False)
        G.add_edge(start_event, converging_gateway)
        G.add_edge(converging_gateway, id(do_part))
        G.add_edge(id(do_part), diverging_gateway)
        G.add_edge(diverging_gateway, end_event)
        G.add_edge(diverging_gateway, id(redo_part))
        G.add_edge(id(redo_part), converging_gateway)

    elif operator == Operator.XOR:
        # One exclusive choice gateway
        exclusive_gateway_diverging = f"ExclusiveGateway_{id(powl_content)}_diverging"
        exclusive_gateway_converging = f"ExclusiveGateway_{id(powl_content)}_converging"
        G.add_node(
            exclusive_gateway_diverging,
            type="diverging",
            paired_with=[exclusive_gateway_converging],
            visited=True,
        )
        G.add_node(
            exclusive_gateway_converging,
            type="converging",
            paired_with=[exclusive_gateway_diverging],
            visited=True,
        )
        G.add_edge(start_event, exclusive_gateway_diverging)
        G.add_edge(exclusive_gateway_converging, end_event)

        for child in powl_content.children:
            if id(child) not in G.nodes:
                G.add_node(id(child), content=child, visited=False)
            G.add_edge(exclusive_gateway_diverging, id(child))
            G.add_edge(id(child), exclusive_gateway_converging)
    else:
        raise ValueError(f"Unsupported operator: {operator}")
    return G


def __handle_decision_graph(powl_content: DecisionGraph) -> nx.DiGraph:
    """
    Handle the DecisionGraph content and return a directed graph.

    Parameters
    ----------
    powl_content : DecisionGraph
        The DecisionGraph content to handle.

    Returns
    -------
    G : nx.DiGraph
        The directed graph representing the DecisionGraph.
    """

    G = nx.DiGraph()
    edges = __obtain_edges(powl_content, transitive_reduction=False)
    node_edges = {
        node: {"incoming": [], "outgoing": []} for node in powl_content.order.nodes
    }
    for edge in edges:
        src, dst = edge
        node_edges[src]["outgoing"].append(dst)
        node_edges[dst]["incoming"].append(src)
    for node in node_edges.keys():
        # Check for end and start
        if type(node) is StartNode:
            G.add_node(id(node), type="start", visited=True)
            # Add an exclusive gateway after it
            G.add_node(
                f"ExclusiveGateway_{id(node)}_afternode",
                type="exclusive_gateway",
                visited=True,
            )
            # Connect them
            G.add_edge(id(node), f"ExclusiveGateway_{id(node)}_afternode")
        elif type(node) is EndNode:
            G.add_node(id(node), type="end", visited=True)
            # Add an exclusive gateway before it
            G.add_node(
                f"ExclusiveGateway_{id(node)}_beforenode",
                type="exclusive_gateway",
                visited=True,
            )
            # Connect them
            G.add_edge(f"ExclusiveGateway_{id(node)}_beforenode", id(node))

    G = __add_auxiliary_nodes_before_after(G, node_edges, type="exclusive_gateway")
    for edge in edges:
        src, dst = edge
        src = (
            f"ExclusiveGateway_{id(src)}_afternode"
            if f"ExclusiveGateway_{id(src)}_afternode" in G.nodes
            else id(src)
        )
        dst = (
            f"ExclusiveGateway_{id(dst)}_beforenode"
            if f"ExclusiveGateway_{id(dst)}_beforenode" in G.nodes
            else id(dst)
        )
        G.add_edge(src, dst)
    return G


def __obtain_edges(powl_content, transitive_reduction: bool = True) -> List[tuple]:
    edges = []
    for src in powl_content.order.nodes:
        for dst in powl_content.order.nodes:
            if powl_content.order.is_edge(src, dst):
                edges.append((src, dst))
    G = nx.DiGraph()
    G.add_edges_from(edges)
    if transitive_reduction:
        G = nx.transitive_reduction(G)
    return list(G.edges)


def __add_auxiliary_nodes_before_after(
    G: nx.DiGraph, neighbors: dict, type: str = "exclusive_gateway"
):
    for node in neighbors.keys():
        # add the node
        if id(node) in G.nodes:
            # It is handled ad-hoc
            continue
        G.add_node(id(node), content=node, visited=False)
        # Add one exclusive gateway before it and connect it

        gateway = (
            f"ExclusiveGateway_{id(node)}_beforenode"
            if type == "exclusive_gateway"
            else f"ParallelGateway_{id(node)}_beforenode"
        )
        if gateway not in G.nodes:
            G.add_node(gateway, type=type, paired_with=[id(node)], visited=True)
        G.add_edge(gateway, id(node))

        # Add one exclusive gateway after it and connect it
        gateway = (
            f"ExclusiveGateway_{id(node)}_afternode"
            if type == "exclusive_gateway"
            else f"ParallelGateway_{id(node)}_afternode"
        )
        if gateway not in G.nodes:
            G.add_node(gateway, type=type, visited=True)
        G.add_edge(id(node), gateway)
    return G


def __handle_StrictPartialOrder(powl_content: StrictPartialOrder) -> nx.DiGraph:
    """
    Handle the StrictPartialOrder content and return a directed graph.

    Parameters
    ----------
    powl_content : StrictPartialOrder
        The StrictPartialOrder content to handle.

    Returns
    -------
    G : nx.DiGraph
        The directed graph representing the StrictPartialOrder.
    """
    edges = __obtain_edges(powl_content)
    G = nx.DiGraph()
    start_event = f"Start_{id(powl_content)}"
    end_event = f"End_{id(powl_content)}"
    G.add_node(start_event, type="start", visited=True)
    G.add_node(end_event, type="end", visited=True)

    # Construct a dictionary with incoming and outgoing edges for each node
    node_edges = {
        node: {"incoming": [], "outgoing": []} for node in powl_content.order.nodes
    }
    for src, dst in edges:
        node_edges[src]["outgoing"].append(dst)
        node_edges[dst]["incoming"].append(src)

    # Preprocess the graph
    G = __add_auxiliary_nodes_before_after(G, node_edges, type="parallel_gateway")
    start_powl = [node for node, edges in node_edges.items() if not edges["incoming"]]
    end_edges = [node for node, edges in node_edges.items() if not edges["outgoing"]]

    # It always has a diverging and converging gateway
    diverging_gateway = f"ParallelGateway_{id(powl_content)}_diverging"
    converging_gateway = f"ParallelGateway_{id(powl_content)}_converging"
    G.add_node(
        diverging_gateway,
        type="diverging",
        paired_with=[converging_gateway],
        visited=True,
    )
    G.add_node(
        converging_gateway,
        type="converging",
        paired_with=[diverging_gateway],
        visited=True,
    )

    # Connect the start and end events to the gateways
    G.add_edge(start_event, diverging_gateway)
    G.add_edge(converging_gateway, end_event)

    # Now, we connect all of the start edges to the diverging gateway
    for start in start_powl:
        gateway_before_node = f"ParallelGateway_{id(start)}_beforenode"
        G.add_edge(diverging_gateway, gateway_before_node)

    # We connect the end events to the converging gateway
    for end in end_edges:
        gateway_after_node = f"ParallelGateway_{id(end)}_afternode"
        G.add_edge(gateway_after_node, converging_gateway)

    for edge in edges:
        src, dst = edge
        after_src = f"ParallelGateway_{id(src)}_afternode"
        before_dst = f"ParallelGateway_{id(dst)}_beforenode"
        G.add_edge(after_src, before_dst)
    return G


def __generate_submodel(powl_content) -> nx.DiGraph:
    """
    Obtain a submodel from the POWL content.

    Parameters
    ----------
    powl_content : POWL
        The POWL content to extract the submodel from.

    Returns
    -------
    G : nx.DiGraph
        The directed graph representing the submodel.
    """
    G = nx.DiGraph()
    handler_map = {
        SilentTransition: __handle_silent_transition,
        Transition: __handle_transition,
        OperatorPOWL: __handle_operator_powl,
        StrictPartialOrder: __handle_StrictPartialOrder,
        DecisionGraph: __handle_decision_graph,
    }

    if type(powl_content) in handler_map:
        try:
            handler = handler_map[type(powl_content)]
        except Exception as e:
            raise ValueError(
                f"Error handling POWL content of type {type(powl_content)}: {e}"
            )
        G = handler(powl_content)
    else:
        raise ValueError(f"Unsupported POWL content type: {type(powl_content)}")
    return G


def __compose_model(
    G: nx.DiGraph, submodel: nx.DiGraph, current_node: nx.Graph
) -> nx.DiGraph:
    """
    Compose two directed graphs into one.

    Parameters
    ----------
    G1 : nx.DiGraph
        The first directed graph.
    G2 : nx.DiGraph
        The second directed graph.
    current_node : nx.Graph
        The current POWL node that is being processed.

    Returns
    -------
    G : nx.DiGraph
        The composed directed graph.
    """
    predecessors = list(G.predecessors(current_node))
    successors = list(G.successors(current_node))

    start_event_nodes = [n for n, deg in submodel.in_degree() if deg == 0]
    end_event_nodes = [n for n, deg in submodel.out_degree() if deg == 0]
    start_event = start_event_nodes[0]
    end_event = end_event_nodes[0]

    # Remove the current node from the original graph
    G.remove_node(current_node)
    start_nodes = list(submodel.successors(start_event))
    end_nodes = list(submodel.predecessors(end_event))
    if start_event is None or end_event is None:
        raise ValueError(
            f"Submodel for {current_node} does not have start or end event."
        )
    submodel.remove_node(start_event)
    submodel.remove_node(end_event)

    # Now, merge the two graphs
    if len(submodel.nodes) == 0:
        # Just connect predecessors and successors
        if len(predecessors) == 1 and len(successors) == 1:
            # We have one-to-one connection, so we can just connect them
            G.add_edge(predecessors[0], successors[0])
            return G
        else:
            raise ValueError(
                f"Unexpected case for node {current_node}: {predecessors} -> {successors}"
            )

    G = nx.compose(G, submodel)
    if len(predecessors) == 1 and len(start_nodes) == 1:
        # We have one-to-one connection, so we can just connect them
        G.add_edge(predecessors[0], start_nodes[0])
        G.add_edge(end_nodes[0], successors[0])
    elif len(predecessors) >= 1 or len(start_nodes) >= 1:
        # We have many-to-many connection, or many-to-one, or one-to-many
        # We have to add a parallel gateway in between
        exclusive_gateway_diverging = (
            f"ExclusiveGateway_{id(current_node)}_pre_additional"
        )
        # Now, we connect all predecessors to this gateway
        G.add_node(exclusive_gateway_diverging, type="exclusive_gateway", visited=True)
        for predecessor in predecessors:
            G.add_edge(predecessor, exclusive_gateway_diverging)
        for start_node in start_nodes:
            G.add_edge(exclusive_gateway_diverging, start_node)
    else:
        raise ValueError(
            f"Unexpected case for node {current_node}: {predecessors} -> {successors}"
        )
    if len(successors) == 1 and len(end_nodes) == 1:
        # We have one-to-one connection, so we can just connect them
        G.add_edge(end_nodes[0], successors[0])
    elif len(successors) >= 1 or len(end_nodes) >= 1:
        # We have many-to-many connection, or many-to-one, or one-to-many
        # We have to add a parallel gateway in between
        exclusive_gateway_converging = (
            f"ExclusiveGateway_{id(current_node)}_post_additional"
        )
        # Now, we connect all successors to this gateway
        G.add_node(exclusive_gateway_converging, type="exclusive_gateway", visited=True)
        for successor in successors:
            G.add_edge(exclusive_gateway_converging, successor)
        for end_node in end_nodes:
            G.add_edge(end_node, exclusive_gateway_converging)
    else:
        raise ValueError(
            f"Unexpected case for node {current_node}: {predecessors} -> {successors}"
        )
    return G


def expand_model(powl, G: nx.DiGraph):
    """
    Recursively expand the POWL model into a directed graph.

    Parameters
    ----------
    powl : POWL
        The POWL model to expand.
    G : nx.DiGraph
        The directed graph to populate.
    """
    # Identify the node that has the powl we are currently considering
    if isinstance(powl, StartNode) or isinstance(powl, EndNode):
        # Remove them from the graph
        node = next((n for n in G.nodes if G.nodes[n].get("content") is powl), None)
        if node is not None:
            G.remove_node(node)
        # And ignore them as they don't add much value
        return G
    node = next((n for n in G.nodes if G.nodes[n].get("content") is powl), None)
    if node is None:
        return G
    submodel = __generate_submodel(powl)
    if submodel is None:
        return G
    G = __compose_model(G, submodel, node)
    nodes_to_handle = [
        node
        for node in submodel.nodes
        if submodel.nodes[node].get("content") is not None
        and submodel.nodes[node]["visited"] is False
    ]
    for node in nodes_to_handle:
        content = submodel.nodes[node]["content"]
        G = expand_model(content, G)
    return G


def __update_paired_with_relation(G: nx.DiGraph, gateway_to_remove, successor):
    """
    Update the paired_with relation for gateways in the graph.

    Parameters
    ----------
    G : nx.DiGraph
        The directed graph.
    gateway_to_remove : str
        The gateway to remove.
    successor : str
        The successor node to connect to the paired gateway.
    """
    current_gateway = G.nodes[gateway_to_remove]
    if "paired_with" in current_gateway:
        paired_with = current_gateway["paired_with"]
        if isinstance(paired_with, list):
            for paired_gateway in paired_with:
                # find the paired gateway in the graph and update its list
                if "Gateway" not in str(paired_gateway):
                    # We only want to update gateways, not tasks
                    continue
                paired_node = G.nodes.get(paired_gateway, None)
                if paired_node is not None:
                    if "paired_with" in paired_node:
                        # If the paired gateway has a list, append the successor
                        if isinstance(paired_node["paired_with"], list):
                            paired_node["paired_with"].append(successor)
                        else:
                            # If it is not a list, convert it to a list
                            paired_node["paired_with"] = [
                                paired_node["paired_with"],
                                successor,
                            ]
                    else:
                        # If it does not have a paired_with attribute, create it
                        paired_node["paired_with"] = [successor]
                    # Remove current gateway from the paired_with list
                    if paired_gateway in paired_node["paired_with"]:
                        paired_node["paired_with"].remove(paired_gateway)
                    # Add it to the successor's paired_with list
                    if "paired_with" in G.nodes[successor]:
                        if isinstance(G.nodes[successor]["paired_with"], list):
                            G.nodes[successor]["paired_with"].append(paired_gateway)
                        else:
                            G.nodes[successor]["paired_with"] = [
                                G.nodes[successor]["paired_with"],
                                paired_gateway,
                            ]
                    else:
                        G.nodes[successor]["paired_with"] = [paired_gateway]
    return G


def __postprocess_graph(G: nx.DiGraph) -> nx.DiGraph:
    G_copy = G.copy()
    while True:
        gateways = [
            node
            for node, _ in G_copy.nodes(data=True)
            if "Parallel" in str(node) or "Exclusive" in str(node)
        ]
        # Get their predecessors and successors
        for gateway in gateways:
            if gateway not in G_copy.nodes:
                continue
            predecessors = list(G_copy.predecessors(gateway))
            successors = list(G_copy.successors(gateway))
            if len(predecessors) == 1 and len(successors) == 1:
                # We can merge them
                G_copy = __update_paired_with_relation(G_copy, gateway, successors[0])
                G_copy.add_edge(predecessors[0], successors[0])
                G_copy.remove_node(gateway)
        # Check if G and G_copy are the same
        if nx.is_isomorphic(G, G_copy):
            break
        G = G_copy.copy()
    return G_copy


def __print_powl(powl):
    print(powl)


def apply(powl):
    """
    Convert a POWL model to a BPMN model.

    Parameters
    ----------
    powl : POWL
        The POWL model to convert.

    Returns
    -------
    bpmn : BPMN
        The converted BPMN model.
    resulting_graph : nx.DiGraph
        The resulting directed graph representing the POWL model.
    original_element_id_to_id : dict
        A mapping from original element IDs to their ides.
    """
    G = nx.DiGraph()
    # Create the start and end event
    start_event = "StartEvent"
    end_event = "EndEvent"
    G.add_node(start_event, type="startEvent", visited=True)
    G.add_node(end_event, type="endEvent", visited=True)
    G.add_node(id(powl), content=powl, visited=False)
    G.add_edge(start_event, id(powl))
    G.add_edge(id(powl), end_event)
    resulting_graph = expand_model(powl, G)
    resulting_graph = __postprocess_graph(resulting_graph)
    try:
        bpmn, original_element_id_to_id = __transform_to_bpmn(resulting_graph)
    except Exception as e:
        raise ValueError(f"Error transforming graph to BPMN: {e}")
    return bpmn, resulting_graph, original_element_id_to_id


def __transform_to_bpmn(G):
    """
    Transform the graph G into a BPMN file.
    """
    # create the root
    node_dict = {node: {"incoming": [], "outgoing": []} for node in G.nodes()}
    node_object_mapping = {}
    original_element_id_to_id = {}
    for u, v, _ in G.edges(data=True):
        node_dict[u]["outgoing"].append(v)
        node_dict[v]["incoming"].append(u)
    bpmn = bpmn_obj.BPMN()

    for node, attrs in G.nodes(data=True):
        object = None
        ided_id = str(id(str(node)))
        if "Start" in str(node):
            ided_id = f"StartEvent_{ided_id}"
            object = bpmn.StartEvent(id=ided_id)

        elif "End" in str(node):
            ided_id = f"EndEvent_{ided_id}"
            object = bpmn.EndEvent(id=ided_id)

        elif "Parallel" in str(node):
            ided_id = f"ParallelGateway_{ided_id}"
            object = bpmn.ParallelGateway(id=ided_id)

        elif "Exclusive" in str(node):
            ided_id = f"ExclusiveGateway_{ided_id}"
            object = bpmn.ExclusiveGateway(id=ided_id)

        elif "Intermediate" in str(node):

            if "Catch" in str(node):
                ided_id = f"IntermediateCatchEvent_{ided_id}"
                object = bpmn.IntermediateCatchEvent(
                    id=ided_id, name=str(attrs.get("content", ""))
                )

            elif "Throw" in str(node):
                ided_id = f"IntermediateThrowEvent_{ided_id}"
                object = bpmn.IntermediateThrowEvent(
                    id=ided_id, name=str(attrs.get("content", ""))
                )
        else:
            # tasks
            ided_id = f"Task_{ided_id}"
            object = bpmn.Task(id=ided_id, name=str(attrs.get("content", "")))
        original_element_id_to_id[str(node)] = ided_id

        node_object_mapping[node] = object
        for incoming in node_dict[node]["incoming"]:
            if incoming not in node_object_mapping:
                # will be handled later
                continue
            seq_flow = bpmn.SequenceFlow(
                source=node_object_mapping[incoming], target=object
            )
            bpmn.add_flow(seq_flow)
            object.add_in_arc(seq_flow)

        for outgoing in node_dict[node]["outgoing"]:
            if outgoing not in node_object_mapping:
                # will be handled later
                continue
            seq_flow = bpmn.SequenceFlow(
                source=object, target=node_object_mapping[outgoing]
            )
            bpmn.add_flow(seq_flow)
            object.add_out_arc(seq_flow)
        bpmn.add_node(object)

    return bpmn, original_element_id_to_id



import copy
from itertools import product
from typing import Hashable, List as TList, Set as TSet, TypeVar

T = TypeVar("T", bound=Hashable)


class BinaryRelation:
    def __init__(self, nodes: TList[T]):
        self._number_nodes = 0
        self._set_nodes(nodes)
        self._edges = [
            [False for _ in range(self._number_nodes)]
            for _ in range(self._number_nodes)
        ]
        self._start_nodes = None
        self._end_nodes = None

    def get_nodes(self) -> TList[T]:
        return self._nodes

    def _set_nodes(self, nodes: TList[T]) -> None:
        self._nodes = [n for n in nodes]
        self._map_node_to_id = {}
        self._map_id_to_node = {}
        n = 0
        for node in self._nodes:
            self._map_node_to_id[node] = n
            self._map_id_to_node[n] = node
            n = n + 1
        self._number_nodes = n

    def add_edge(self, source: T, target: T) -> None:
        try:
            i = self._map_node_to_id[source]
            j = self._map_node_to_id[target]
        except Exception:
            raise Exception("Unable to create edge! Invalid  source or target!")
        else:
            self._edges[i][j] = True

    def remove_edge(self, source: T, target: T) -> None:
        try:
            i = self._map_node_to_id[source]
            j = self._map_node_to_id[target]
        except Exception:
            raise Exception("Unable to remove edge! Invalid  source or target!")
        else:
            self._edges[i][j] = False

    def remove_edge_without_violating_transitivity(self, source: T, target: T) -> None:
        try:
            i = self._map_node_to_id[source]
            j = self._map_node_to_id[target]
        except Exception:
            raise Exception("Unable to remove edge! Invalid  source or target!")
        else:
            self._edges[i][j] = False
            n = len(self.nodes)
            changed = True
            while changed:
                changed = False
                for i, j, k in product(range(n), range(n), range(n)):
                    if (
                        i != j
                        and j != k
                        and self._edges[i][j]
                        and self._edges[j][k]
                        and not self._edges[i][k]
                    ):
                        self._edges[j][k] = False
                        changed = True

    def add_node(self, node: T) -> None:
        if node not in self._nodes:
            self._nodes.append(node)
            n = self._number_nodes
            self._map_node_to_id[node] = n
            self._map_id_to_node[n] = node
            new_edges = [[False for _ in range(n + 1)] for _ in range(n + 1)]
            for i in range(n):
                for j in range(n):
                    new_edges[i][j] = self._edges[i][j]

            self._edges = new_edges
            self._number_nodes = n + 1

    def is_edge(self, source, target) -> bool:
        try:
            i = self._map_node_to_id[source]
            j = self._map_node_to_id[target]
        except Exception:
            print(source)
            print(target)
            print(self._nodes)
            raise Exception("Unable to create edge! Invalid  source or target!")
        else:
            return self._edges[i][j]

    def is_edge_id(self, i: int, j: int) -> bool:
        return self._edges[i][j]

    def get_transitive_reduction(self) -> "BinaryRelation":
        if not self.is_irreflexive():
            raise ValueError(
                "Cannot generate transitive reduction! Reflexivity detected!"
            )

        tc = BinaryRelation(self.nodes)
        tc._edges = copy.deepcopy(self._edges)
        tc.add_transitive_edges()

        tr = BinaryRelation(self.nodes)
        tr._edges = copy.deepcopy(self._edges)

        n = len(self.nodes)
        for i, j, k in product(range(n), range(n), range(n)):
            if (
                i != j
                and j != k
                and tc.edges[i][j]
                and tc.edges[j][k]
                and tc.edges[i][k]
            ):
                tr._edges[i][k] = False
        return tr

    def add_transitive_edges(self) -> None:
        n = len(self.nodes)
        changed = True
        while changed:
            changed = False
            for i, j, k in product(range(n), range(n), range(n)):
                if (
                    i != j
                    and j != k
                    and self._edges[i][j]
                    and self._edges[j][k]
                    and not self.is_edge_id(i, k)
                ):
                    self._edges[i][k] = True
                    changed = True

    def is_strict_partial_order(self) -> bool:
        return self.is_irreflexive() and self.is_transitive()

    def get_start_nodes(self) -> TSet[T]:
        return {
            self._map_id_to_node[j]
            for j in range(len(self.nodes))
            if not any(self._edges[i][j] for i in range(len(self.nodes)))
        }

    def get_end_nodes(self) -> TSet[T]:
        return {
            self._map_id_to_node[i]
            for i in range(len(self.nodes))
            if not any(self._edges[i][j] for j in range(len(self.nodes)))
        }

    def is_irreflexive(self) -> bool:
        n = len(self.nodes)
        for i in range(n):
            if self.is_edge_id(i, i):
                return False
        return True

    def is_transitive(self) -> bool:
        n = len(self.nodes)
        for i, j, k in product(range(n), range(n), range(n)):
            if (
                self.is_edge_id(i, j)
                and self.is_edge_id(j, k)
                and not self.is_edge_id(i, k)
            ):
                return False
        return True

    def __repr__(self) -> str:
        res = "(nodes = {  "
        for node in self._nodes:
            res = res + node.__repr__() + ", "
        res = res[:-2]
        res = res + "  }, order = {  "
        for node in self._nodes:
            i = self._map_node_to_id[node]
            for node2 in self._nodes:
                j = self._map_node_to_id[node2]
                if self._edges[i][j]:
                    res = res + node.__repr__() + "-->" + node2.__repr__() + ", "
        res = res[:-2]
        return res + "  })"

    nodes = property(get_nodes, _set_nodes)

    @property
    def edges(self):
        return self._edges

    def get_preset(self, child):
        res = set()
        for node in self.nodes:
            if self.is_edge(node, child):
                res.add(node)
        return res

    def get_postset(self, child):
        res = set()
        for node in self.nodes:
            if self.is_edge(child, node):
                res.add(node)
        return res



from powl.main import (
    convert_from_workflow_net,
    convert_to_bpmn,
    convert_to_petri_net,
    discover,
    discover_from_dfg,
    discover_from_partially_ordered_log,
    discover_petri_net_from_ocel,
    import_event_log,
    import_ocel,
    save_visualization,
    save_visualization_net,
    view,
    view_net,
)

__name__ = "powl"
__version__ = "2.2.9"



import importlib
import tempfile
from enum import Enum
from typing import Any, Dict, Optional

import graphviz
from graphviz import Digraph
from pm4py.objects.bpmn.obj import BPMN

from pm4py.util import constants, exec_utils
from pm4py.visualization.common import gview, save as gsave


SPLIT_LABELS = False


class Parameters(Enum):
    FORMAT = "format"
    RANKDIR = "rankdir"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"


def get_label(name):
    parts = name.split(" ")
    if SPLIT_LABELS and len(parts) >= 3:
        return " ".join(parts[:2]) + "\n" + " ".join(parts[2:])
    else:
        return name


def apply(
    bpmn_graph: BPMN, parameters: Optional[Dict[Any, Any]] = None
) -> graphviz.Digraph:
    """
    Visualize a BPMN graph

    Parameters
    -------------
    bpmn_graph
        BPMN graph
    parameters
        Parameters of the visualization, including:
         - Parameters.FORMAT: the format of the visualization
         - Parameters.RANKDIR: the direction of the representation (default: LR)

    Returns
    ------------
    gviz
        Graphviz representation
    """
    if parameters is None:
        parameters = {}

    from pm4py.objects.bpmn.obj import BPMN
    from pm4py.objects.bpmn.util.sorting import get_sorted_nodes_edges

    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 28)
    font_size = str(font_size)
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        "", filename=filename.name, engine="dot", graph_attr={"bgcolor": bgcolor}
    )
    viz.graph_attr["rankdir"] = rankdir
    # viz.attr(nodesep='1')
    # viz.attr(ranksep='0.4')

    nodes, edges = get_sorted_nodes_edges(bpmn_graph)


    for n in nodes:
        n_id = str(id(n))
        if isinstance(n, BPMN.StartEvent):

            viz.node(
                n_id,
                label="",
                shape="circle",
                fontsize=font_size,
                width="0.6",
                height="0.6",
            )
        elif isinstance(n, BPMN.EndEvent):

            viz.node(
                n_id,
                label="",
                shape="circle",
                fontsize=font_size,
                penwidth="3.0",
                width="0.6",
                height="0.6",
            )
        elif isinstance(n, BPMN.ParallelGateway):
            with importlib.resources.path(
                "powl.visualization.powl.variants.icons", "gate_and.svg"
            ) as gimg:
                xor_image = str(gimg)
                viz.node(
                    n_id,
                    label="",
                    shape="diamond",
                    width="0.8",
                    height="0.8",
                    fixedsize="true",
                    image=xor_image,
                )
            # viz.node(n_id, label="+", shape="diamond", fontsize=font_size)
        elif isinstance(n, BPMN.ExclusiveGateway):

            with importlib.resources.path(
                "powl.visualization.powl.variants.icons", "gate.svg"
            ) as gimg:
                xor_image = str(gimg)
                viz.node(
                    n_id,
                    label="",
                    shape="diamond",
                    width="0.8",
                    height="0.8",
                    fixedsize="true",
                    image=xor_image,
                )
            # viz.node(n_id, label="X", shape="diamond", fontsize=font_size)
        elif isinstance(n, BPMN.InclusiveGateway):
            viz.node(n_id, label="O", shape="diamond", fontsize=font_size)
        elif isinstance(n, BPMN.Task):
            viz.node(
                n_id, shape="box", label=get_label(n.get_name()), fontsize=font_size
            )
        else:
            viz.node(n_id, label="", shape="circle", fontsize=font_size)

    for e in edges:
        n_id_1 = str(id(e[0]))
        n_id_2 = str(id(e[1]))

        viz.edge(n_id_1, n_id_2, penwidth="2.0")


    viz.attr(overlap="false")

    viz.format = image_format.replace("html", "plain-ext")

    return viz


def save(gviz: graphviz.Digraph, output_file_path: str, parameters=None):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    gsave.save(gviz, output_file_path, parameters=parameters)
    return ""


def view(gviz: graphviz.Digraph, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    if True:
        return gview.view(gviz, parameters=parameters)


def matplotlib_view(gviz: graphviz.Digraph, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """
    if True:
        return gview.matplotlib_view(gviz, parameters=parameters)



from copy import copy
from itertools import combinations

from pm4py import PetriNet
from pm4py.algo.analysis.workflow_net import algorithm as wf_eval
from pm4py.objects.petri_net.utils import petri_utils as pn_util

from powl.conversion.to_powl.from_pn.utils.subnet_creation import (
    add_arc_from_to,
    id_generator, clone_place,
)


def validate_workflow_net(net: PetriNet):
    places_no_incoming = [p for p in net.places if not p.in_arcs]
    if len(places_no_incoming) == 1:
        start_place = places_no_incoming[0]
    else:
        raise Exception(f"Not a WF-net!")

    places_no_outgoing = [p for p in net.places if not p.out_arcs]
    if len(places_no_outgoing) == 1:
        end_place = places_no_outgoing[0]
    else:
        raise Exception(f"Not a WF-net!")

    if not wf_eval.apply(net):
        raise Exception(f"Not a WF-net!")

    return start_place, end_place


def remove_initial_and_end_silent_activities(
    net: PetriNet, start_places: set[PetriNet.Place], end_places: set[PetriNet.Place]
):
    change = True
    while change and len(net.transitions) > 1:
        change = False
        if len(start_places) == 1:
            start_place = list(start_places)[0]
            if len(start_place.in_arcs) == 0 and len(start_place.out_arcs) == 1:
                transition = list(start_place.out_arcs)[0].target
                if len(transition.in_arcs) == 1 and is_silent(transition):
                    pn_util.remove_place(net, start_place)
                    start_places.remove(start_place)
                    next_places = list(pn_util.post_set(transition))
                    pn_util.remove_transition(net, transition)
                    for p in next_places:
                        start_places.add(p)
                    change = True

    change = True
    while change and len(net.transitions) > 1:
        change = False
        if len(end_places) == 1:
            end_place = list(end_places)[0]
            if len(end_place.in_arcs) == 1 and len(end_place.out_arcs) == 0:
                transition = list(end_place.in_arcs)[0].source
                if len(transition.out_arcs) == 1 and is_silent(transition):
                    pn_util.remove_transition(net, transition)
                    pn_util.remove_place(net, end_place)
                    end_places.remove(end_place)
                    prev_places = list(pn_util.pre_set(transition))
                    for p in prev_places:
                        end_places.add(p)
                    change = True

    return start_places, end_places


def __get_identical_place(place: PetriNet.Place, places_set: set[PetriNet.Place]):
    for other in places_set:
        if pn_util.post_set(place) == pn_util.post_set(other) and pn_util.pre_set(
            place
        ) == pn_util.pre_set(other):
            return other
    return None


def __remove_and_replace_if_present(
    old_p: PetriNet.Place, new_p: PetriNet.Place, place_set: set[PetriNet.Place]
):
    if old_p in place_set:
        place_set.remove(old_p)
        if new_p not in place_set:
            place_set.add(new_p)
    return place_set


def remove_duplicated_places(
    net: PetriNet, start_places: set[PetriNet.Place], end_places: set[PetriNet.Place]
):
    all_places = list(net.places)
    places_to_keep = {all_places[0]}
    for place in all_places[1:]:
        other = __get_identical_place(place, places_to_keep)
        if other:
            pn_util.remove_place(net, place)
            start_places = __remove_and_replace_if_present(place, other, start_places)
            end_places = __remove_and_replace_if_present(place, other, end_places)
        else:
            places_to_keep.add(place)

    return start_places, end_places


def remove_unconnected_places(
    net: PetriNet, start_places: set[PetriNet.Place], end_places: set[PetriNet.Place]
):
    places = list(net.places)
    for p in places:
        if len(p.in_arcs) == 0 == len(p.out_arcs):
            pn_util.remove_place(net, p)
            start_places.remove(p)
            end_places.remove(p)

    return start_places, end_places


def preprocess(net):
    all_places = net.places
    for p1, p2 in combinations(all_places, 2):
        pre1 = pn_util.pre_set(p1)
        pre2 = pn_util.pre_set(p2)
        post1 = pn_util.post_set(p1)
        post2 = pn_util.post_set(p2)

        if (pre1 == pre2) and (post1 == post2):
            pn_util.remove_place(net, p2)
            return preprocess(net)

        if pre1 == pre2:
            common_post = post1 & post2
            if len(pre1) > 1 or len(common_post) > 0:
                new_place = PetriNet.Place(f"place_{next(id_generator())}")
                net.places.add(new_place)

                for transition in pre1:
                    add_arc_from_to(transition, new_place, net)
                    arcs_to_remove = p1.in_arcs | p2.in_arcs
                    for arc in arcs_to_remove:
                        pn_util.remove_arc(net, arc)

                for transition in common_post:
                    add_arc_from_to(new_place, transition, net)
                    out_arcs = p1.out_arcs | p2.out_arcs
                    for arc in out_arcs:
                        if arc.target in common_post:
                            pn_util.remove_arc(net, arc)

                new_silent = PetriNet.Transition(
                    f"silent_transition_{next(id_generator())}"
                )
                net.transitions.add(new_silent)
                add_arc_from_to(new_place, new_silent, net)
                add_arc_from_to(new_silent, p1, net)
                add_arc_from_to(new_silent, p2, net)
                return preprocess(net)

        if post1 == post2:
            common_pre = pre1 & pre2
            if len(post1) > 1 or len(common_pre) > 0:
                new_place = PetriNet.Place(f"place_{next(id_generator())}")
                net.places.add(new_place)

                for transition in post1:
                    add_arc_from_to(new_place, transition, net)
                    arcs_to_remove = p1.out_arcs | p2.out_arcs
                    for arc in arcs_to_remove:
                        pn_util.remove_arc(net, arc)

                for transition in common_pre:
                    add_arc_from_to(transition, new_place, net)
                    in_arcs = p1.in_arcs | p2.in_arcs
                    for arc in in_arcs:
                        if arc.source in common_pre:
                            pn_util.remove_arc(net, arc)

                new_silent = PetriNet.Transition(
                    f"silent_transition_{next(id_generator())}"
                )
                net.transitions.add(new_silent)
                add_arc_from_to(p1, new_silent, net)
                add_arc_from_to(p2, new_silent, net)
                add_arc_from_to(new_silent, new_place, net)

                return preprocess(net)

    return net


def add_new_start_and_end_if_needed(
    net, start_places: set[PetriNet.Place], end_places: set[PetriNet.Place]
):
    if len(start_places) == 0 or len(end_places) == 0:
        raise Exception("This should not happen!")

    if len(start_places) > 1:

        new_source_id = f"source_{next(id_generator())}"
        new_source = __redirect_shared_arcs_to_new_place(
            net, list(start_places), new_source_id
        )

        if new_source:

            new_silent = PetriNet.Transition(f"silent_start_{next(id_generator())}")
            net.transitions.add(new_silent)

            for p in start_places:
                add_arc_from_to(new_silent, p, net)

            add_arc_from_to(new_source, new_silent, net)
            start_places = {new_source}

    if len(end_places) > 1:

        new_sink_id = f"sink_{next(id_generator())}"
        new_sink = __redirect_shared_arcs_to_new_place(
            net, list(end_places), new_sink_id
        )

        if new_sink:

            new_silent = PetriNet.Transition(f"silent_end_{next(id_generator())}")
            net.transitions.add(new_silent)

            for p in end_places:
                add_arc_from_to(p, new_silent, net)

            add_arc_from_to(new_silent, new_sink, net)

            end_places = {new_sink}

    return start_places, end_places


def __redirect_shared_arcs_to_new_place(
    net, places: list[PetriNet.Place], new_place_id
):
    shared_pre_set = set(pn_util.pre_set(places[0]))
    for p in places[1:]:
        shared_pre_set &= set(pn_util.pre_set(p))

    shared_post_set = set(pn_util.post_set(places[0]))
    for p in places[1:]:
        shared_post_set &= set(pn_util.post_set(p))

    arcs = list(net.arcs)

    for arc in arcs:
        source = arc.source
        target = arc.target
        if (source in shared_pre_set and target in places) or (
            source in places and target in shared_post_set
        ):
            pn_util.remove_arc(net, arc)

    if len(shared_post_set) > 0 or len(shared_pre_set) > 0:
        new_place = PetriNet.Place(new_place_id)
        net.places.add(new_place)

        for node in shared_pre_set:
            add_arc_from_to(node, new_place, net)

        for node in shared_post_set:
            add_arc_from_to(new_place, node, net)

        return new_place

    else:
        return None


def is_silent(transition) -> bool:
    return transition.label is None


def make_self_loop_explicit(net: PetriNet, start_place: PetriNet.Place, end_place: PetriNet.Place):
    if start_place == end_place:
        place = start_place
        place_copy = clone_place(net, place, {})
        out_arcs = place.out_arcs
        for arc in list(out_arcs):
            target = arc.target
            pn_util.remove_arc(net, arc)
            add_arc_from_to(place_copy, target, net)
        do_transition = PetriNet.Transition(f"silent_do_{place.name}", None)
        do = set()
        do.add(do_transition)
        net.transitions.add(do_transition)
        add_arc_from_to(place, do_transition, net)
        add_arc_from_to(do_transition, place_copy, net)
        return net, place, place_copy
    else:
        return net, start_place, end_place


from typing import Any, Dict, List, Optional, Tuple

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough, T

from powl.discovery.total_order_based.inductive.variants.decision_graph.dfg_cut import (
    DFGCutUVCL,
)
from powl.objects.obj import DecisionGraph, POWL


class DFGFallThroughUVCL(FallThrough[IMDataStructureUVCL]):
    @classmethod
    def holds(
        cls,
        obj: IMDataStructureUVCL,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> bool:
        return True

    @classmethod
    def apply(
        cls,
        obj: T,
        pool=None,
        manager=None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[DecisionGraph, List[POWL]]]:
        return DFGCutUVCL.apply(obj, parameters)



# this file was initially copied from https://github.com/Nik314/DF2-Miner


def get_interaction_patterns(relations):

    convergent_object_types = {a: set() for a in relations["ocel:activity"].unique()}
    divergent_object_types = {a: set() for a in relations["ocel:activity"].unique()}
    deficient_object_types = {a: set() for a in relations["ocel:activity"].unique()}
    related_object_types = {
        a: set(relations["ocel:type"].unique())
        for a in relations["ocel:activity"].unique()
    }

    look_up_dict_activities = relations.set_index("ocel:eid").to_dict()["ocel:activity"]
    look_up_dict_objects = relations.set_index("ocel:oid").to_dict()["ocel:type"]

    identifiers = (
        relations.groupby("ocel:eid")
        .apply(
            lambda frame: tuple(sorted(set(frame["ocel:oid"].values))),
            include_groups=False,
        )
        .to_frame(name="all")
    )
    identifiers["activity"] = [
        look_up_dict_activities[event_id] for event_id in identifiers.index
    ]

    for activity in relations["ocel:activity"].unique():
        sub_relations = relations[relations["ocel:activity"] == activity]
        for object_type in relations["ocel:type"].unique():
            sub_sub_relations = sub_relations[sub_relations["ocel:type"] == object_type]
            if (
                sub_sub_relations["ocel:eid"].nunique()
                != sub_relations["ocel:eid"].nunique()
            ):
                if not sub_sub_relations["ocel:eid"].nunique() > 0:
                    related_object_types[activity].remove(object_type)
                else:
                    if (
                        sub_sub_relations["ocel:eid"].nunique()
                        < sub_relations["ocel:eid"].nunique()
                    ):
                        deficient_object_types[activity].add(object_type)

    for object_type in relations["ocel:type"].unique():
        identifiers[object_type] = identifiers["all"].apply(
            lambda object_set: tuple(
                sorted(
                    list(
                        {
                            object_id
                            for object_id in object_set
                            if look_up_dict_objects[object_id] == object_type
                        }
                    )
                )
            )
        )

    for object_type in relations["ocel:type"].unique():
        sub_identifiers = identifiers[identifiers[object_type] != set()]
        for activity in relations["ocel:activity"].unique():
            sub_sub_identifiers = sub_identifiers[
                sub_identifiers["activity"] == activity
            ]

            matches = sub_sub_identifiers.groupby(object_type).apply(
                lambda frame: frame["all"].nunique(), include_groups=False
            )
            matches = matches[[index for index in matches.index if index]]

            if (
                sub_sub_identifiers[object_type]
                .apply(lambda object_set: len(object_set))
                .max()
                > 1
            ):
                convergent_object_types[activity].add(object_type)
            if matches.max() > 1:
                divergent_object_types[activity].add(object_type)

    return (
        divergent_object_types,
        convergent_object_types,
        related_object_types,
        deficient_object_types,
    )



from abc import ABC
from collections import Counter
from itertools import combinations
from typing import Any, Collection, Dict, List, Optional

from pm4py.algo.discovery.inductive.cuts import utils as cut_util
from pm4py.algo.discovery.inductive.cuts.abc import T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL
from pm4py.util import exec_utils
from pm4py.algo.discovery.inductive.variants.imf import IMFParameters

from powl.discovery.total_order_based.inductive.utils.filtering import FILTERING_TYPE, FilteringType
from powl.discovery.total_order_based.inductive.variants.decision_graph.max_decision_graph_cut import (
    MaximalDecisionGraphCut
)


class CyclicDecisionGraphCut(MaximalDecisionGraphCut[T], ABC):
    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Any]]:

        dfg = obj.dfg
        alphabet = parameters["alphabet"]

        groups = [frozenset([a]) for a in alphabet]

        for a, b in combinations(alphabet, 2):
            if (a, b) in dfg.graph and (b, a) in dfg.graph:
                groups = cut_util.merge_groups_based_on_activities(a, b, groups)

        if len(groups) < 2:
            return None

        return groups


class CyclicDecisionGraphCutUVCL(CyclicDecisionGraphCut[IMDataStructureUVCL], ABC):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:

        filtering = False
        if FILTERING_TYPE in parameters.keys():
            filtering_type = parameters[FILTERING_TYPE]
            if filtering_type is FilteringType.DFG_FREQUENCY:
                noise_threshold = exec_utils.get_param_value(
                    IMFParameters.NOISE_THRESHOLD, parameters, 0.0
                )
                if noise_threshold > 0.0:
                    filtering = True

        logs = [Counter() for _ in groups]
        for t, freq in obj.data_structure.items():
            for i, group in enumerate(groups):
                seg = []
                last = None
                for e in t:
                    if e in group:
                        if len(seg) > 0 and last not in group:
                            if not filtering or obj.dfg.graph.get((last, e), 0) > obj.dfg.graph.get((seg[-1], e), 0):
                                logs[i][tuple(seg)] += freq
                                seg = []
                        seg.append(e)
                    last = e
                if len(seg) > 0:
                    logs[i][tuple(seg)] += freq

        return [IMDataStructureUVCL(l) for l in logs]


from abc import ABC
from itertools import combinations
from typing import Any, Collection, Dict, List, Optional

from pm4py.algo.discovery.inductive.cuts import utils as cut_util
from pm4py.objects.dfg import util as dfu
from pm4py.algo.discovery.inductive.cuts.abc import T
from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureUVCL

from powl.discovery.total_order_based.inductive.variants.decision_graph.cyclic_dg_cut import (
    CyclicDecisionGraphCut,
    CyclicDecisionGraphCutUVCL,
)


class StrictCyclicDecisionGraphCut(CyclicDecisionGraphCut[T], ABC):
    @classmethod
    def holds(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[List[Any]]:

        dfg = obj.dfg
        alphabet = parameters["alphabet"]
        groups = [frozenset([a]) for a in alphabet]

        def _get_group(activity):
            for g in groups:
                if activity in g:
                    return g
            raise Exception(f"activity {activity} not in {groups}")

        changed = True

        while changed:
            changed = False
            for (a, b), (c, d) in combinations(dfg.graph, 2):
                if {a, b, c, d}.issubset(alphabet):
                    if (
                        _get_group(a) != _get_group(b)
                        and _get_group(a) == _get_group(d)
                        and _get_group(b) == _get_group(c)
                    ):
                        groups = cut_util.merge_groups_based_on_activities(a, b, groups)
                        changed = True

        if len(groups) < 2:
            return None

        return groups


class StrictCyclicDecisionGraphCutUVCL(
    StrictCyclicDecisionGraphCut[IMDataStructureUVCL], ABC
):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureUVCL,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureUVCL]:

        return CyclicDecisionGraphCutUVCL.project(obj, groups, parameters)



import importlib.resources
import tempfile
from enum import Enum

from graphviz import Digraph
from pm4py.objects.process_tree.obj import Operator
from pm4py.util import exec_utils

from powl.objects.obj import (
    DecisionGraph,
    EndNode,
    FrequentTransition,
    OperatorPOWL,
    POWL,
    SilentTransition,
    StartNode,
    StrictPartialOrder,
    Transition,
)

min_width = "1.5"  # Set the minimum width in inches
fillcolor = "#fcfcfc"
opacity_change_ratio = 0.02
FONT_SIZE = "18"
PEN_WIDTH = "1"


class Parameters(Enum):
    FORMAT = "format"
    COLOR_MAP = "color_map"
    ENABLE_DEEPCOPY = "enable_deepcopy"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"


def apply(powl: POWL) -> Digraph:
    """
    Obtain a POWL model representation through GraphViz

    Parameters
    -----------
    powl
        POWL model

    Returns
    -----------
    gviz
        GraphViz Digraph
    """

    filename = tempfile.NamedTemporaryFile(suffix=".gv")

    viz = Digraph("powl", filename=filename.name, engine="dot")
    viz.attr("node", shape="ellipse", fixedsize="false")
    viz.attr(nodesep="1")
    viz.attr(ranksep="1")
    viz.attr(compound="true")
    viz.attr(overlap="scale")
    viz.attr(splines="true")
    viz.attr(rankdir="TB")
    viz.attr(style="filled")
    viz.attr(fillcolor=fillcolor)

    color_map = exec_utils.get_param_value(Parameters.COLOR_MAP, {}, {})

    repr_powl(powl, viz, color_map, level=0, skip_order=False, loop_order=False)
    viz.format = "svg"

    return viz


def get_color(node, color_map):
    """
    Gets a color for a node from the color map

    Parameters
    --------------
    node
        Node
    color_map
        Color map
    """
    if node in color_map:
        return color_map[node]
    return "black"


def get_block_id(powl):
    if (
        isinstance(powl, OperatorPOWL)
        or isinstance(powl, StrictPartialOrder)
        or isinstance(powl, DecisionGraph)
    ):
        return "cluster_" + str(id(powl))
    else:
        raise Exception(f"Unknown POWL type {type(powl)}!")


def add_operator_edge(vis, current_node_id, child_ids, directory="none", style=""):
    child_base_id, block_id = child_ids
    if block_id:
        vis.edge(
            current_node_id,
            child_base_id,
            dir=directory,
            lhead=block_id,
            style=style,
            minlen="2",
            penwidth=PEN_WIDTH,
        )
    else:
        vis.edge(
            current_node_id,
            child_base_id,
            dir=directory,
            style=style,
            penwidth=PEN_WIDTH,
        )


def add_order_edge(
    block, child_1_ids, child_2_ids, directory="forward", color="black", style=""
):
    child_1_base_id, block_id_1 = child_1_ids
    child_2_base_id, block_id_2 = child_2_ids

    if block_id_1:
        if block_id_2:
            block.edge(
                child_1_base_id,
                child_2_base_id,
                dir=directory,
                color=color,
                style=style,
                ltail=block_id_1,
                lhead=block_id_2,
                minlen="2",
                penwidth=PEN_WIDTH,
            )
        else:
            block.edge(
                child_1_base_id,
                child_2_base_id,
                dir=directory,
                color=color,
                style=style,
                ltail=block_id_1,
                minlen="2",
                penwidth=PEN_WIDTH,
            )
    else:
        if block_id_2:
            block.edge(
                child_1_base_id,
                child_2_base_id,
                dir=directory,
                color=color,
                style=style,
                lhead=block_id_2,
                minlen="2",
                penwidth=PEN_WIDTH,
            )
        else:
            block.edge(
                child_1_base_id,
                child_2_base_id,
                dir=directory,
                color=color,
                style=style,
                penwidth=PEN_WIDTH,
            )


def mark_block(block, skip_order, loop_order):
    if skip_order:
        if loop_order:
            with importlib.resources.path(
                "pm4py.visualization.powl.variants.icons", "skip-loop-tag.svg"
            ) as gimg:
                image = str(gimg)
                block.attr(
                    label=f"""<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0">
                                        <TR><TD WIDTH="55" HEIGHT="27" FIXEDSIZE="TRUE"><IMG SRC="{image}" SCALE="BOTH"/></TD></TR>
                                        </TABLE>>"""
                )
                block.attr(labeljust="r")
        else:
            with importlib.resources.path(
                "pm4py.visualization.powl.variants.icons", "skip-tag.svg"
            ) as gimg:
                image = str(gimg)
                block.attr(
                    label=f"""<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0">
                                        <TR><TD WIDTH="55" HEIGHT="27" FIXEDSIZE="TRUE"><IMG SRC="{image}" SCALE="BOTH"/></TD></TR>
                                        </TABLE>>"""
                )
                block.attr(labeljust="r")
    elif loop_order:
        with importlib.resources.path(
            "pm4py.visualization.powl.variants.icons", "loop-tag.svg"
        ) as gimg:
            image = str(gimg)
            block.attr(
                label=f"""<<TABLE BORDER="0" CELLBORDER="0" CELLSPACING="0">
                                    <TR><TD WIDTH="55" HEIGHT="27" FIXEDSIZE="TRUE"><IMG SRC="{image}" SCALE="BOTH"/></TD></TR>
                                    </TABLE>>"""
            )
            block.attr(labeljust="r")
    else:
        block.attr(label="")


def repr_powl(powl, viz, color_map, level, skip_order, loop_order):
    font_size = FONT_SIZE
    this_node_id = str(id(powl))

    current_color = darken_color(fillcolor, amount=opacity_change_ratio * level)
    block_id = None

    if isinstance(powl, FrequentTransition) or (isinstance(powl, Transition) and (skip_order or loop_order)):
        if isinstance(powl, FrequentTransition):
            skip_order = skip_order or powl.skippable
            loop_order = loop_order or powl.selfloop
            label = powl.activity
        else:
            label = powl.label
        if skip_order:
            if loop_order:
                with importlib.resources.path(
                    "powl.visualization.powl.variants.icons", "skip-loop-tag.svg"
                ) as gimg:
                    image = str(gimg)
                    viz.node(
                        this_node_id,
                        label="\n" + label,
                        imagepos="tr",
                        image=image,
                        shape="box",
                        width=min_width,
                        fontsize=font_size,
                        style="filled",
                        fillcolor=current_color,
                    )
            else:
                with importlib.resources.path(
                    "powl.visualization.powl.variants.icons", "skip-tag.svg"
                ) as gimg:
                    image = str(gimg)
                    viz.node(
                        this_node_id,
                        label="\n" + label,
                        imagepos="tr",
                        image=image,
                        shape="box",
                        width=min_width,
                        fontsize=font_size,
                        style="filled",
                        fillcolor=current_color,
                    )
        else:
            if loop_order:
                with importlib.resources.path(
                    "powl.visualization.powl.variants.icons", "loop-tag.svg"
                ) as gimg:
                    image = str(gimg)
                    viz.node(
                        this_node_id,
                        label="\n" + label,
                        imagepos="tr",
                        image=image,
                        shape="box",
                        width=min_width,
                        fontsize=font_size,
                        style="filled",
                        fillcolor=current_color,
                    )
            else:
                viz.node(
                    this_node_id,
                    label=label,
                    shape="box",
                    width=min_width,
                    fontsize=font_size,
                    style="filled",
                    fillcolor=current_color,
                )

    elif isinstance(powl, Transition):
        if isinstance(powl, SilentTransition):
            viz.node(
                this_node_id,
                label="",
                style="filled",
                fillcolor="black",
                shape="square",
                width="0.3",
                height="0.3",
                fixedsize="true",
            )
        else:
            label = f"<{str(powl.label)}"
            if powl._role is not None and powl._organization is not None:
                # Add a label to the box
                label += f"""<br/><font color="grey" point-size="10">({powl._organization}, {powl._role})</font><br/>"""
            elif powl._organization is not None:
                # Add a label to the box
                label += f"""<br/><font color="grey" point-size="10">({powl._organization})</font><br/>"""
            elif powl._role is not None:
                # Add a label to the box
                label += f"""<br/><font color="grey" point-size="10">({powl._role})</font><br/>"""

            label += ">"

            viz.node(
                this_node_id,
                label,
                shape="box",
                fontsize=font_size,
                width=min_width,
                style="filled",
                fillcolor=current_color,
            )

    elif isinstance(powl, StrictPartialOrder):
        transitive_reduction = powl.order.get_transitive_reduction()
        block_id = get_block_id(powl)
        child_id_map = {}
        with viz.subgraph(name=block_id) as block:
            block.attr(margin="20,20")
            block.attr(style="filled")
            block.attr(fillcolor=current_color)

            mark_block(block, skip_order, loop_order)

            this_node_id = make_anchor(block, block_id)

            for child in powl.children:
                child_id_map[child] = repr_powl(
                    child,
                    block,
                    color_map,
                    level=level + 1,
                    skip_order=False,
                    loop_order=False,
                )
            for child in powl.children:
                for child2 in powl.children:
                    if transitive_reduction.is_edge(child, child2):
                        add_order_edge(block, child_id_map[child], child_id_map[child2])

    elif isinstance(powl, StartNode) or isinstance(powl, EndNode):
        with importlib.resources.path(
            "powl.visualization.powl.variants.icons", "gate_navy.svg"
        ) as gimg:
            xor_image = str(gimg)
            viz.node(
                this_node_id,
                label="",
                shape="diamond",
                color="navy",
                width="0.6",
                height="0.6",
                fixedsize="true",
                image=xor_image,
            )
    elif isinstance(powl, DecisionGraph):
        if len(powl.children) == 1:
            child = powl.children[0]
            if (
                isinstance(child, StrictPartialOrder)
                or isinstance(child, DecisionGraph)
                or isinstance(child, OperatorPOWL)
            ):
                if powl.order.is_edge(powl.start, powl.end):
                    skip_order = True
                if powl.order.is_edge(child, child):
                    loop_order = True

                return repr_powl(
                    child,
                    viz,
                    color_map,
                    level=level,
                    skip_order=skip_order,
                    loop_order=loop_order,
                )

        block_id = get_block_id(powl)
        child_id_map = {}
        with viz.subgraph(name=block_id) as block:
            block.attr(margin="20,20")
            block.attr(style="filled")
            block.attr(fillcolor=current_color)

            mark_block(block, skip_order, loop_order)

            this_node_id = make_anchor(block, block_id)

            for child in powl.order.nodes:
                loop_child = powl.order.is_edge(child, child)
                child_id_map[child] = repr_powl(
                    child,
                    block,
                    color_map,
                    level=level + 1,
                    skip_order=False,
                    loop_order=loop_child,
                )

            for child in powl.order.nodes:
                for child2 in powl.order.nodes:
                    if child != child2 and powl.order.is_edge(child, child2):
                        add_order_edge(
                            block,
                            child_id_map[child],
                            child_id_map[child2],
                            color="navy",
                            style="dashed",
                        )

    elif isinstance(powl, OperatorPOWL):
        if powl.operator == Operator.XOR:
            silent_children = [
                child for child in powl.children if isinstance(child, SilentTransition)
            ]
            if len(silent_children) > 0:
                other_children = [
                    child
                    for child in powl.children
                    if not isinstance(child, SilentTransition)
                ]
                if len(other_children) == 1:
                    child = other_children[0]
                    if (
                        isinstance(child, StrictPartialOrder)
                        or isinstance(child, DecisionGraph)
                        or isinstance(child, OperatorPOWL)
                    ):
                        return repr_powl(
                            child,
                            viz,
                            color_map,
                            level=level,
                            skip_order=True,
                            loop_order=loop_order,
                        )
                elif len(other_children) > 1:
                    children_powl = OperatorPOWL(
                        operator=powl.operator, children=other_children
                    )
                    return repr_powl(
                        children_powl,
                        viz,
                        color_map,
                        level=level,
                        skip_order=True,
                        loop_order=loop_order,
                    )

        if powl.operator == Operator.LOOP:
            do = powl.children[0]
            redo = powl.children[1]
            if isinstance(do, SilentTransition) and (
                isinstance(redo, StrictPartialOrder)
                or isinstance(redo, DecisionGraph)
                or isinstance(redo, OperatorPOWL)
            ):
                return repr_powl(
                    redo, viz, color_map, level=level, skip_order=True, loop_order=True
                )
            if isinstance(redo, SilentTransition) and (
                isinstance(do, StrictPartialOrder)
                or isinstance(do, DecisionGraph)
                or isinstance(do, OperatorPOWL)
            ):
                return repr_powl(
                    do,
                    viz,
                    color_map,
                    level=level,
                    loop_order=True,
                    skip_order=skip_order,
                )

        block_id = get_block_id(powl)
        with viz.subgraph(name=block_id) as block:
            block.attr(margin="20,20")
            block.attr(style="filled")
            block.attr(fillcolor=current_color)
            mark_block(block, skip_order, loop_order)
            if powl.operator == Operator.LOOP:
                with importlib.resources.path(
                    "powl.visualization.powl.variants.icons", "loop.svg"
                ) as gimg:
                    image = str(gimg)
                    block.node(
                        this_node_id,
                        image=image,
                        label="",
                        fontsize=font_size,
                        width="0.4",
                        height="0.4",
                        fixedsize="true",
                    )
                    anchor_id = make_anchor(block, block_id)
                do = powl.children[0]
                redo = powl.children[1]
                do_id = repr_powl(
                    do,
                    block,
                    color_map,
                    level=level + 1,
                    skip_order=False,
                    loop_order=False,
                )
                add_operator_edge(block, this_node_id, do_id)
                redo_id = repr_powl(
                    redo,
                    block,
                    color_map,
                    level=level + 1,
                    skip_order=False,
                    loop_order=False,
                )
                add_operator_edge(block, this_node_id, redo_id, style="dashed")
            elif powl.operator == Operator.XOR:
                with importlib.resources.path(
                    "powl.visualization.powl.variants.icons", "xor.svg"
                ) as gimg:
                    image = str(gimg)
                    block.node(
                        this_node_id,
                        image=image,
                        label="",
                        fontsize=font_size,
                        width="0.4",
                        height="0.4",
                        fixedsize="true",
                    )
                    anchor_id = make_anchor(block, block_id)
                for child in powl.children:
                    child_id = repr_powl(
                        child,
                        block,
                        color_map,
                        level=level + 1,
                        skip_order=False,
                        loop_order=False,
                    )
                    add_operator_edge(block, this_node_id, child_id)
            else:
                raise NotImplementedError
            this_node_id = anchor_id
    else:
        raise Exception(f"Unknown POWL operator: {type(powl)}")

    return this_node_id, block_id


def darken_color(color, amount):
    """Darkens the given color by the specified amount"""
    import matplotlib.colors as mcolors

    amount = min(0.3, amount)

    rgb = mcolors.to_rgb(color)
    darker = [x * (1 - amount) for x in rgb]
    return mcolors.to_hex(darker)


def make_anchor(block, block_id):
    anchor_id = f"anchor_{block_id}"
    block.node(
        anchor_id,
        label="",
        shape="point",
        width="0.01",
        height="0.01",
        fixedsize="true",
        style="invis",
    )
    return anchor_id



from abc import ABC
from collections import Counter
from typing import Any, Collection, Dict, Generic, List, Optional, Tuple

from pm4py.algo.discovery.inductive.base_case.abc import T
from pm4py.algo.discovery.inductive.cuts.sequence import (
    SequenceCut,
    SequenceCutDFG,
    SequenceCutUVCL,
    StrictSequenceCut,
    StrictSequenceCutDFG,
    StrictSequenceCutUVCL,
)
from pm4py.algo.discovery.inductive.dtypes.im_dfg import InductiveDFG
from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG

from powl.objects.obj import Sequence


class POWLSequenceCut(SequenceCut, ABC, Generic[T]):
    @classmethod
    def operator(cls, parameters: Optional[Dict[str, Any]] = None) -> Sequence:
        return Sequence([])

    @classmethod
    def apply(
        cls, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[Sequence, List[T]]]:
        g = cls.holds(obj, parameters)
        if g is None:
            return g
        children = cls.project(obj, g, parameters)
        po = Sequence(children)
        return po, children


class POWLStrictSequenceCut(POWLSequenceCut[T], StrictSequenceCut, ABC):
    pass


class POWLSequenceCutUVCL(SequenceCutUVCL, POWLSequenceCut[IMDataStructureUVCL]):
    pass


class POWLStrictSequenceCutUVCL(
    StrictSequenceCutUVCL, StrictSequenceCut[IMDataStructureUVCL], POWLSequenceCutUVCL
):
    pass


class POWLSequenceCutDFG(SequenceCutDFG, POWLSequenceCut[IMDataStructureDFG]):
    @classmethod
    def project(
        cls,
        obj: IMDataStructureDFG,
        groups: List[Collection[Any]],
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[IMDataStructureDFG]:
        dfg = obj.dfg
        start_activities = []
        end_activities = []
        activities = []
        dfgs = []
        skippable = []
        for g in groups:
            skippable.append(False)
        activities_idx = {}
        for gind, g in enumerate(groups):
            for act in g:
                activities_idx[act] = int(gind)
        i = 0
        while i < len(groups):
            to_succ_arcs = Counter()
            from_prev_arcs = Counter()
            if i < len(groups) - 1:
                for (a, b) in dfg.graph:
                    if a in groups[i] and b in groups[i + 1]:
                        to_succ_arcs[a] += dfg.graph[(a, b)]

            if i > 0:
                for (a, b) in dfg.graph:
                    if a in groups[i - 1] and b in groups[i]:
                        from_prev_arcs[b] += dfg.graph[(a, b)]

            if i == 0:
                start_activities.append({})
                for a in dfg.start_activities:
                    if a in groups[i]:
                        start_activities[i][a] = dfg.start_activities[a]
                    else:
                        j = i
                        while j < activities_idx[a]:
                            skippable[j] = True
                            j = j + 1
            else:
                start_activities.append(from_prev_arcs)

            if i == len(groups) - 1:
                end_activities.append({})
                for a in dfg.end_activities:
                    if a in groups[i]:
                        end_activities[i][a] = dfg.end_activities[a]
                    else:
                        j = activities_idx[a] + 1
                        while j <= i:
                            skippable[j] = True
                            j = j + 1
            else:
                end_activities.append(to_succ_arcs)

            activities.append({})
            act_count = dfu.get_vertex_frequencies(dfg)
            for a in groups[i]:
                activities[i][a] = act_count[a]
            dfgs.append({})
            for (a, b) in dfg.graph:
                if a in groups[i] and b in groups[i]:
                    dfgs[i][(a, b)] = dfg.graph[(a, b)]
            i = i + 1
        i = 0
        while i < len(dfgs):
            dfi = DFG()
            [dfi.graph.update({(a, b): dfgs[i][(a, b)]}) for (a, b) in dfgs[i]]
            [
                dfi.start_activities.update({a: start_activities[i][a]})
                for a in start_activities[i]
            ]
            [
                dfi.end_activities.update({a: end_activities[i][a]})
                for a in end_activities[i]
            ]
            dfgs[i] = dfi
            i = i + 1
        for (a, b) in dfg.graph:
            z = activities_idx[b]
            j = activities_idx[a] + 1
            while j < z:
                skippable[j] = True
                j = j + 1

        return [
            IMDataStructureDFG(InductiveDFG(dfg=dfgs[i], skip=skippable[i]))
            for i in range(len(dfgs))
        ]


class POWLStrictSequenceCutDFG(
    StrictSequenceCutDFG, StrictSequenceCut[IMDataStructureDFG], POWLSequenceCutDFG
):
    pass



from multiprocessing import Manager, Pool
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

from pm4py.algo.discovery.inductive.cuts.abc import Cut, T

from pm4py.algo.discovery.inductive.dtypes.im_ds import (
    IMDataStructureDFG,
    IMDataStructureUVCL,
)
from pm4py.algo.discovery.inductive.fall_through.abc import FallThrough

from powl.discovery.total_order_based.inductive.fall_through.activity_concurrent import (
    POWLActivityConcurrentUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.activity_once_per_trace import (
    POWLActivityOncePerTraceUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.decision_graph.dfg_fall_through import (
    DFGFallThroughUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.flower import (
    POWLFlowerModelDFG,
    POWLFlowerModelUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.strict_tau_loop import (
    POWLStrictTauLoopUVCL,
)
from powl.discovery.total_order_based.inductive.fall_through.tau_loop import (
    POWLTauLoopUVCL,
)
from powl.objects.obj import POWL

S = TypeVar("S", bound=FallThrough | Cut)


class FallThroughFactory:
    @classmethod
    def get_fall_throughs(
        cls,
        obj: T,
        enable_dfg_fall_through,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> List[Type[S]]:
        if type(obj) is IMDataStructureUVCL:
            if enable_dfg_fall_through:
                return [
                    POWLActivityOncePerTraceUVCL,
                    POWLActivityConcurrentUVCL,
                    POWLStrictTauLoopUVCL,
                    POWLTauLoopUVCL,
                    DFGFallThroughUVCL,
                ]
            else:
                return [
                    POWLActivityOncePerTraceUVCL,
                    POWLActivityConcurrentUVCL,
                    POWLStrictTauLoopUVCL,
                    POWLTauLoopUVCL,
                    POWLFlowerModelUVCL,
                ]
        elif type(obj) is IMDataStructureDFG:
            return [POWLFlowerModelDFG]
        return list()

    @classmethod
    def fall_through(
        cls,
        obj: T,
        pool: Pool,
        manager: Manager,
        enable_dfg_fall_through,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[POWL, List[T]]]:
        for f in FallThroughFactory.get_fall_throughs(obj, enable_dfg_fall_through):
            r = f.apply(obj, pool, manager, parameters)
            if r is not None:
                return r
        return None



from enum import Enum
from typing import Any, Dict, Optional

import graphviz
from pm4py.objects.bpmn.obj import BPMN
from pm4py.util import exec_utils
from pm4py.visualization.bpmn.variants import dagrejs

from powl.visualization.bpmn.variants import classic


class Variants(Enum):
    CLASSIC = classic
    DAGREJS = dagrejs


DEFAULT_VARIANT = Variants.CLASSIC


def apply(
    bpmn_graph: BPMN,
    variant=DEFAULT_VARIANT,
    parameters: Optional[Dict[Any, Any]] = None,
) -> graphviz.Digraph:
    """
    Visualize a BPMN graph

    Parameters
    -------------
    bpmn_graph
        BPMN graph
    variant
        Variant of the visualization, possible values:
         - Variants.CLASSIC
    parameters
        Version-specific parameters

    Returns
    ------------
    gviz
        Graphviz representation
    """
    return exec_utils.get_variant(variant).apply(bpmn_graph, parameters=parameters)


def save(
    gviz: graphviz.Digraph,
    output_file_path: str,
    variant=DEFAULT_VARIANT,
    parameters=None,
):
    """
    Save the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    output_file_path
        Path where the GraphViz output should be saved
    """
    return exec_utils.get_variant(variant).save(
        gviz, output_file_path, parameters=parameters
    )


def view(gviz: graphviz.Digraph, variant=DEFAULT_VARIANT, parameters=None):
    """
    View the diagram

    Parameters
    -----------
    gviz
        GraphViz diagram
    """
    return exec_utils.get_variant(variant).view(gviz, parameters=parameters)


def matplotlib_view(gviz: graphviz.Digraph, variant=DEFAULT_VARIANT, parameters=None):
    """
    Views the diagram using Matplotlib

    Parameters
    ---------------
    gviz
        Graphviz
    """
    return exec_utils.get_variant(variant).matplotlib_view(gviz, parameters=parameters)



from pm4py.objects.petri_net.obj import Marking, PetriNet

from powl.conversion.utils.pn_reduction import add_arc_from_to
from powl.objects.BinaryRelation import BinaryRelation
from powl.objects.obj import (
    DecisionGraph,
    Operator,
    OperatorPOWL,
    Sequence,
    SilentTransition,
    StrictPartialOrder,
    Transition,
)
from powl.objects.oc_powl import ComplexModel, LeafNode, ObjectCentricPOWL


DIV_IF_ALL_CHILDREN_DIV = True


def generate_xor(children):
    if len(children) == 1:
        return children[0]
    rel = BinaryRelation(children)
    return DecisionGraph(rel, children, children)


def generate_flower_model(children):
    xor = generate_xor(children)
    return OperatorPOWL(operator=Operator.LOOP, children=[SilentTransition(), xor])


def clone_workflow_net(
    net: PetriNet,
    im: Marking,
    fm: Marking,
    name_suffix: str = "_copy",
    label_delimiter: str = "<|>",
):
    """
    Create a deep copy of a Petri net and its initial/final markings.
    - Appends `name_suffix` to place/transition names.
    - If a transition label contains `label_delimiter`, keeps only the part before it.
    """
    mapping = {}

    new_net = PetriNet(f"{net.name}{name_suffix}")

    for place in net.places:
        new_place = PetriNet.Place(f"{place.name}{name_suffix}")
        new_net.places.add(new_place)
        mapping[place] = new_place

    for t in net.transitions:
        label = t.label
        if label and label_delimiter in label:
            label = label.split(label_delimiter, 1)[0].strip()
        new_t = PetriNet.Transition(name=f"{t.name}{name_suffix}", label=label)
        new_net.transitions.add(new_t)
        mapping[t] = new_t

    for arc in net.arcs:
        add_arc_from_to(mapping[arc.source], mapping[arc.target], new_net)

    new_im = Marking({mapping[p]: im[p] for p in im})
    new_fm = Marking({mapping[p]: fm[p] for p in fm})

    return new_net, new_im, new_fm


def project_oc_powl(oc_powl: ObjectCentricPOWL, object_type, div_edges):

    if isinstance(oc_powl, LeafNode):
        if oc_powl.activity == "" or object_type not in oc_powl.related:
            return SilentTransition()
        activity = oc_powl.activity
        if object_type in oc_powl.get_type_information()[(activity, "div")]:
            return generate_flower_model([Transition(label=activity)])
        return Transition(label=activity)

    assert isinstance(oc_powl, ComplexModel)
    related_activities = set(
        [
            a
            for a in oc_powl.get_activities()
            if object_type in oc_powl.get_type_information()[(a, "rel")] and a != ""
        ]
    )

    if not related_activities:
        return SilentTransition()

    if all(
        object_type in oc_powl.get_type_information()[(a, "div")]
        for a in related_activities
    ):
        return generate_flower_model([Transition(label=a) for a in related_activities])

    else:
        loop = (
            isinstance(oc_powl.flat_model, OperatorPOWL)
            and oc_powl.flat_model.operator == Operator.LOOP
        )
        parallel = (
            isinstance(oc_powl.flat_model, StrictPartialOrder)
            and len(oc_powl.flat_model.order.edges) == 0
        )
        if loop:
            return OperatorPOWL(
                operator=Operator.LOOP,
                children=[
                    project_oc_powl(sub, object_type, div_edges)
                    for sub in oc_powl.oc_children
                ],
            )
        if parallel:
            return StrictPartialOrder(
                [
                    project_oc_powl(sub, object_type, div_edges)
                    for sub in oc_powl.oc_children
                ]
            )

        diverging = [
            i
            for i in range(len(oc_powl.oc_children))
            if oc_powl.oc_children[i].get_activities() & related_activities
            and all(
                object_type in oc_powl.get_type_information()[(a, "div")]
                for a in oc_powl.oc_children[i].get_activities() & related_activities
            )
        ]
        non_diverging = [
            i
            for i in range(len(oc_powl.oc_children))
            if oc_powl.oc_children[i].get_activities() & related_activities
            and i not in diverging
        ]

        if isinstance(oc_powl.flat_model, StrictPartialOrder):
            div_activities = set(
                sum(
                    [
                        list(
                            oc_powl.oc_children[i].get_activities() & related_activities
                        )
                        for i in diverging
                    ],
                    [],
                )
            )
            div_activities = {a for a in div_activities if a != ""}

            if div_activities:
                div_subtree = generate_flower_model(
                    [Transition(label=a) for a in div_activities]
                )
                if len(non_diverging) > 0:
                    mapping = {
                        oc_powl.flat_model.children[i]: (
                            project_oc_powl(
                                oc_powl.oc_children[i], object_type, div_edges
                            )
                            if i in non_diverging
                            else SilentTransition()
                        )
                        for i in range(len(oc_powl.oc_children))
                    }
                    non_div_subtree = oc_powl.flat_model.map_nodes(mapping)
                    return StrictPartialOrder(nodes=[non_div_subtree, div_subtree])
                else:
                    return div_subtree
            else:
                mapping = {
                    oc_powl.flat_model.children[i]: project_oc_powl(
                        oc_powl.oc_children[i], object_type, div_edges
                    )
                    for i in range(len(oc_powl.oc_children))
                }
                return oc_powl.flat_model.map_nodes(mapping)

        elif isinstance(oc_powl.flat_model, DecisionGraph):

            if DIV_IF_ALL_CHILDREN_DIV:
                parts = _partition_children(
                    oc_powl, diverging, related_activities, div_edges
                )

                mapping = {}
                processed_ids = set()
                for group in parts:
                    div_children = [
                        Transition(a)
                        for i in group
                        for a in oc_powl.oc_children[i].get_activities()
                        & related_activities
                    ]
                    flower = generate_flower_model(div_children)
                    for i in group:
                        processed_ids.add(i)
                        mapping[oc_powl.flat_model.children[i]] = flower

                for i in range(len(oc_powl.oc_children)):
                    if i in processed_ids:
                        continue
                    mapping[oc_powl.flat_model.children[i]] = project_oc_powl(
                        oc_powl.oc_children[i], object_type, div_edges
                    )

                return oc_powl.flat_model.map_nodes(mapping)

            else:
                parts = _partition_children(
                    oc_powl, diverging, related_activities, div_edges
                )

                mapping = {}

                for group in parts:
                    if len(group) == 1:
                        i = list(group)[0]
                        mapping[oc_powl.flat_model.children[i]] = project_oc_powl(
                            oc_powl.oc_children[i], object_type, div_edges
                        )
                    else:
                        div_children = [
                            project_oc_powl(
                                oc_powl.oc_children[i], object_type, div_edges
                            )
                            for i in group
                        ]
                        flower = generate_flower_model(div_children)
                        for i in group:
                            mapping[oc_powl.flat_model.children[i]] = flower

                return oc_powl.flat_model.map_nodes(mapping)

        else:
            raise NotImplementedError


def _partition_children(oc_powl, diverging, related_activities, div_edges):
    edges = [tuple(edge) for edge in div_edges]

    if DIV_IF_ALL_CHILDREN_DIV:
        parts = [{i} for i in diverging]
    else:
        parts = [{i} for i in range(len(oc_powl.oc_children))]

    def find_group(a):
        for g in parts:
            for i in g:
                if a in oc_powl.oc_children[i].get_activities():
                    return g
        return None

    for u, v in edges:
        if u in related_activities and v in related_activities:
            gu = find_group(u)
            gv = find_group(v)
            if gu is None or gv is None:
                continue
            if gu is gv:
                continue
            gu |= gv
            parts.remove(gv)

    return parts


def handle_deficiency(oc_powl: ObjectCentricPOWL):

    if isinstance(oc_powl, LeafNode):
        if oc_powl.activity == "":
            return oc_powl, []
        else:
            from itertools import chain, combinations

            stable_types = oc_powl.related - oc_powl.deficient
            variable_types = oc_powl.related & oc_powl.deficient
            if variable_types:
                ot_sets = [
                    stable_types | {c for c in comb}
                    for comb in chain.from_iterable(
                        combinations(variable_types, n)
                        for n in range(len(variable_types) + 1)
                    )
                ]

                mapping = {}
                for ots in ot_sets:
                    transition = Transition(
                        oc_powl.activity + "<|>" + str(sorted(list(ots)))
                    )
                    mapping[transition] = LeafNode(
                        transition=transition,
                        related=ots,
                        convergent=oc_powl.convergent & ots,
                        deficient=set(),
                        divergent=oc_powl.divergent & ots,
                    )
                flat_model = generate_xor(children=list(mapping.keys()))
                return ComplexModel(flat_model=flat_model, mapping=mapping), [
                    oc_powl.activity
                ]
            else:
                return oc_powl, []

    assert isinstance(oc_powl, ComplexModel)
    sub_results = [handle_deficiency(sub) for sub in oc_powl.oc_children]
    flat_children = [child[0].flat_model for child in sub_results]
    flat_to_oc_mapping = {
        flat_children[i]: sub_results[i][0] for i in range(len(sub_results))
    }
    old_flat_to_new_flat_mapping = {
        oc_powl.flat_model.children[i]: flat_children[i]
        for i in range(len(oc_powl.flat_model.children))
    }

    flat_model = oc_powl.flat_model

    if isinstance(flat_model, Sequence):
        new_flat_model = Sequence(flat_children)
    elif isinstance(flat_model, StrictPartialOrder) or isinstance(
        flat_model, DecisionGraph
    ):
        new_flat_model = flat_model.map_nodes(old_flat_to_new_flat_mapping)
    elif isinstance(flat_model, OperatorPOWL):
        new_flat_model = OperatorPOWL(
            operator=flat_model.operator, children=flat_children
        )
    else:
        raise NotImplementedError

    return ComplexModel(new_flat_model, flat_to_oc_mapping), sum(
        [sub[1] for sub in sub_results], []
    )


def convert_ocpowl_to_ocpn(oc_powl: ObjectCentricPOWL, divergence_matrices):

    assert isinstance(oc_powl, ObjectCentricPOWL)
    nets = {}
    nets_duplicates = {}

    convergent_activities = {}
    activities = set()
    oc_powl, special_activities = handle_deficiency(oc_powl)

    for ot in oc_powl.get_object_types():
        powl_model = project_oc_powl(oc_powl, ot, divergence_matrices[ot])
        powl_model = powl_model.reduce_silent_transitions(add_empty_paths=False)
        powl_model = powl_model.simplify()
        from powl.conversion.converter import apply as to_pn

        net, im, fm = to_pn(powl_model)
        nets[ot] = net, im, fm
        nets_duplicates[ot] = clone_workflow_net(net, im, fm, label_delimiter="<|>")
        activities.update(
            {
                a
                for a in oc_powl.get_activities()
                if ot in oc_powl.get_type_information()[(a, "rel")]
            }
        )
        convergent_activities[ot] = {
            a: ot in oc_powl.get_type_information()[(a, "con")]
            for a in oc_powl.get_activities()
        }

    ocpn = {
        "activities": activities,
        "object_types": nets.keys(),
        "petri_nets": nets,
        "petri_nets_with_duplicates": nets_duplicates,
        "double_arcs_on_activity": convergent_activities,
        "tbr_results": {},
    }

    return ocpn



from collections import deque

from pm4py.objects.petri_net.obj import PetriNet
from pm4py.objects.petri_net.utils import petri_utils as pn_util


def get_simplified_reachability_graph(net: PetriNet):
    graph = {node: set() for node in net.transitions}
    for start_node in graph.keys():
        reachable = set()
        queue = deque()
        queue.append(start_node)
        while queue:
            node = queue.popleft()
            if node not in reachable:
                reachable.add(node)
                successors = pn_util.post_set(node)
                queue.extend(successors)
        graph[start_node].update(reachable)
    return graph


def get_reachable_without_looping(start_node, stop_node):
    reachable = set()
    queue = deque()
    queue.append(start_node)
    while queue:
        node = queue.popleft()
        if node != stop_node and node not in reachable:
            reachable.add(node)
            successors = pn_util.post_set(node)
            queue.extend(successors)
    return reachable


def get_reachable_transitions_from_place_to_another(
    start_place: PetriNet.Place, end_place: PetriNet.Place
):
    visited = set()
    queue = deque()
    queue.append(start_place)
    while queue:
        node = queue.popleft()
        if node not in visited:
            visited.add(node)
            if node == end_place:
                continue
            successors = pn_util.post_set(node)
            queue.extend(successors)
    visited = {node for node in visited if isinstance(node, PetriNet.Transition)}
    return visited


from typing import Any, Dict, List, Optional, Tuple, TypeVar

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructureLog

from powl.discovery.total_order_based.inductive.variants.dynamic_clustering.factory import (
    CutFactoryPOWLDynamicClustering,
)
from powl.discovery.total_order_based.inductive.variants.im_tree import IMBasePOWL
from powl.discovery.total_order_based.inductive.variants.powl_discovery_varaints import (
    POWLDiscoveryVariant,
)
from powl.objects.obj import POWL

T = TypeVar("T", bound=IMDataStructureLog)


class POWLInductiveMinerDynamicClustering(IMBasePOWL):
    def instance(self) -> POWLDiscoveryVariant:
        return POWLDiscoveryVariant.DYNAMIC_CLUSTERING

    def find_cut(
        self, obj: T, parameters: Optional[Dict[str, Any]] = None
    ) -> Optional[Tuple[POWL, List[T]]]:
        res = CutFactoryPOWLDynamicClustering.find_cut(obj, parameters=parameters)
        return res



import tempfile
from collections import Counter
from copy import copy
from enum import Enum
from typing import Any, Dict, Optional

import graphviz

from graphviz import Digraph
from pm4py.objects.dfg import util as dfu
from pm4py.objects.dfg.obj import DFG

from pm4py.util import constants, exec_utils


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    FORMAT = "format"
    MAX_NO_EDGES_IN_DIAGRAM = "maxNoOfEdgesInDiagram"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    FONT_SIZE = "font_size"
    RANKDIR = "rankdir"
    BGCOLOR = "bgcolor"


def apply(
    dfg_obj: DFG, parameters: Optional[Dict[Any, Any]] = None
) -> graphviz.Digraph:
    """
    Visualize a frequency directly-follows graph

    Parameters
    -----------------
    dfg_obj
        Directly-follows graph
    parameters
        Variant-specific parameters

    Returns
    -----------------
    gviz
        Graphviz digraph
    """
    if parameters is None:
        parameters = {}

    max_no_of_edges_in_diagram = exec_utils.get_param_value(
        Parameters.MAX_NO_EDGES_IN_DIAGRAM, parameters, 100000
    )
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 32)
    font_size = str(font_size)

    dfg = dfg_obj.graph
    start_activities = dfg_obj.start_activities
    end_activities = dfg_obj.end_activities

    activities = dfu.get_vertices(dfg_obj)

    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )

    activities_count = Counter({key: 0 for key in activities})
    for el in dfg:
        activities_count[el[1]] += dfg[el]
    if isinstance(start_activities, dict):
        for act in start_activities:
            activities_count[act] += start_activities[act]

    return graphviz_visualization(
        activities_count,
        dfg,
        measure="frequency",
        max_no_of_edges_in_diagram=max_no_of_edges_in_diagram,
        start_activities=start_activities,
        end_activities=end_activities,
        font_size=font_size,
        bgcolor=bgcolor,
        rankdir=rankdir,
    )


def graphviz_visualization(
    activities_count,
    dfg,
    start_activities,
    end_activities,
    measure="frequency",
    max_no_of_edges_in_diagram=100000,
    font_size="12",
    bgcolor=constants.DEFAULT_BGCOLOR,
    rankdir=constants.DEFAULT_RANKDIR_GVIZ,
):
    """
    Do GraphViz visualization of a DFG graph

    Parameters
    -----------
    activities_count
        Count of attributes in the log (may include attributes that are not in the DFG graph)
    dfg
        DFG graph
    measure
        Describes which measure is assigned to edges in direcly follows graph (frequency/performance)
    max_no_of_edges_in_diagram
        Maximum number of edges in the diagram allowed for visualization
    start_activities
        Start activities of the log
    end_activities
        End activities of the log
    font_size
        Size of the text on the activities/edges
    bgcolor
        Background color of the visualization (i.e., 'transparent', 'white', ...)
    rankdir
        Direction of the graph ("LR" for left-to-right; "TB" for top-to-bottom)

    Returns
    -----------
    viz
        Digraph object
    """

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        "",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor, "rankdir": rankdir},
    )

    # first, remove edges in diagram that exceeds the maximum number of edges in the diagram
    dfg_key_value_list = []
    for edge in dfg:
        dfg_key_value_list.append([edge, dfg[edge]])
    # more fine grained sorting to avoid that edges that are below the threshold are
    # undeterministically removed
    dfg_key_value_list = sorted(
        dfg_key_value_list, key=lambda x: (x[1], x[0][0], x[0][1]), reverse=True
    )
    dfg_key_value_list = dfg_key_value_list[
        0 : min(len(dfg_key_value_list), max_no_of_edges_in_diagram)
    ]
    dfg_allowed_keys = [x[0] for x in dfg_key_value_list]
    dfg_keys = list(dfg.keys())
    for edge in dfg_keys:
        if edge not in dfg_allowed_keys:
            del dfg[edge]

    activities_count_int = copy(activities_count)

    activities_in_dfg = set(activities_count)

    # represent nodes
    viz.attr("node", shape="box")

    if len(activities_in_dfg) == 0:
        activities_to_include = sorted(list(set(activities_count_int)))
    else:
        # take unique elements as a list not as a set (in this way, nodes are added in the same order to the graph)
        activities_to_include = sorted(list(set(activities_in_dfg)))

    activities_map = {}

    for act in activities_to_include:
        viz.node(str(hash(act)), act, fontsize=font_size)
        activities_map[act] = str(hash(act))

    # make edges addition always in the same order
    dfg_edges = sorted(list(dfg.keys()))

    # represent edges
    for edge in dfg_edges:
        label = str(dfg[edge])
        viz.edge(
            str(hash(edge[0])), str(hash(edge[1])), fontsize=font_size, penwidth="2.0"
        )

    start_activities_to_include = [
        act for act in start_activities if act in activities_map
    ]
    end_activities_to_include = [act for act in end_activities if act in activities_map]

    if start_activities_to_include:
        viz.node(
            "@@startnode",
            "START",
            style="filled",
            fillcolor="lightgrey",
            fontsize=font_size,
        )
        for act in start_activities_to_include:
            label = (
                str(start_activities[act])
                if isinstance(start_activities, dict) and measure == "frequency"
                else ""
            )
            viz.edge(
                "@@startnode", activities_map[act], fontsize=font_size, penwidth="2.0"
            )

    if end_activities_to_include:
        # <&#9632;>
        viz.node(
            "@@endnode",
            "END",
            style="filled",
            fillcolor="lightgrey",
            fontsize=font_size,
        )
        for act in end_activities_to_include:
            label = (
                str(end_activities[act])
                if isinstance(end_activities, dict) and measure == "frequency"
                else ""
            )
            viz.edge(
                activities_map[act], "@@endnode", fontsize=font_size, penwidth="2.0"
            )

    viz.attr(overlap="false")
    viz.attr(fontsize="11")

    viz.format = "SVG"

    return viz
